import pytest
from testutils.text import dedent_multiline


# Override default element to take four optional arguments.
@pytest.fixture
def fake_element_cls(fake_element_cls):
    class FakeElement(fake_element_cls):
        def __init__(self, name, a=None, b=None, c=None, d=None):
            super().__init__(name)
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    return FakeElement


@pytest.fixture
def spec(spec, fake_element_adapter_factory, fake_element_cls):
    spec.register_element(fake_element_adapter_factory(fake_element_cls))
    return spec


@pytest.mark.parametrize(
    "script,graph_definition",
    (
        pytest.param(
            # Element with no arguments.
            "fake_element myelement",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label "kat"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label "kat.0"
                        token "'fake_element'"
                        name_token "'myelement'"
                        type "ELEMENT"
                        extra_tokens "[' ']"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                ]
                """
            ),
            id="element with no arguments",
        ),
        pytest.param(
            "fake_element myelement 1 2 c=3 d=4",
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label "kat"
                        type "ROOT"
                        extra_tokens "[]"
                    ]
                    node [
                        id 1
                        label "kat.0"
                        token "'fake_element'"
                        name_token "'myelement'"
                        type "ELEMENT"
                        extra_tokens "[' ', ' ', ' ', ' ', ' ']"
                    ]
                    node [
                        id 2
                        label "kat.0.0"
                        token "'1'"
                        type "VALUE"
                    ]
                    node [
                        id 3
                        label "kat.0.1"
                        token "'2'"
                        type "VALUE"
                    ]
                    node [
                        id 4
                        label "kat.0.2"
                        token "'3'"
                        key_token "'c'"
                        equals_token "'='"
                        type "VALUE"
                    ]
                    node [
                        id 5
                        label "kat.0.3"
                        token "'4'"
                        key_token "'d'"
                        equals_token "'='"
                        type "VALUE"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 1
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 3
                        target 1
                        type "ARGUMENT"
                        order "1"
                    ]
                    edge [
                        source 4
                        target 1
                        type "ARGUMENT"
                        order "2"
                    ]
                    edge [
                        source 5
                        target 1
                        type "ARGUMENT"
                        order "3"
                    ]
                ]
                """
            ),
            id="element with arguments",
        ),
        pytest.param(
            dedent_multiline(
                """
                fake_element myelement1
                fake_element myelement2
                """
            ),
            dedent_multiline(
                """
                graph [
                    directed 1
                    node [
                        id 0
                        label "kat"
                        type "ROOT"
                        extra_tokens "['\\n']"
                    ]
                    node [
                        id 1
                        label "kat.0"
                        token "'fake_element'"
                        name_token "'myelement1'"
                        type "ELEMENT"
                        extra_tokens "[' ']"
                    ]
                    node [
                        id 2
                        label "kat.1"
                        token "'fake_element'"
                        name_token "'myelement2'"
                        type "ELEMENT"
                        extra_tokens "[' ']"
                    ]
                    edge [
                        source 1
                        target 0
                        type "ARGUMENT"
                        order "0"
                    ]
                    edge [
                        source 2
                        target 0
                        type "ARGUMENT"
                        order "1"
                    ]
                ]
                """
            ),
            id="multiple elements",
        ),
    ),
)
def test_element(assert_graphs_match, script, graph_definition):
    assert_graphs_match(script, graph_definition)



"""Collection of Actions that deal with sensing tasks such as computing sensing
matrices, optimising RF readouts, etc."""

import logging

import numpy as np

from finesse.components import DegreeOfFreedom
from finesse.components.readout import ReadoutDetectorOutput, ReadoutRF
from finesse.exceptions import FinesseException
from finesse.parameter import Parameter, ParameterRef
from finesse.utilities.misc import deprecation_warning

from ...parameter import deref
from ...solutions import BaseSolution
from ...utilities.tables import NumberTable
from ...simulations.sparse.simulation import SparseMatrixSimulation
from .utils import elements_to_name
from .base import Action
from .lti import FrequencyResponse
from finesse.utilities import OrderedSet

LOGGER = logging.getLogger(__name__)


def get_readout_workspace(readouts, output, readout_workspaces):
    """Return the readout workspaces for a particular readout's quadrature outputs.

    Parameters
    ----------
    readouts : iterable
        :class:`.ReadoutRF` elements
    output : str
        Quadrature output to try and select, 'I', 'Q', 'DC'
    readout_workspaces : dict
        Workspaces of the readouts in the simulation
    """
    return tuple(readout_workspaces[rd.name + "_" + output] for rd in readouts)


class OptimiseRFReadoutPhaseDCSolution(BaseSolution):
    pass


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class OptimiseRFReadoutPhaseDC(Action):
    """This optimises the demodulation phase of :class:`.ReadoutRF` elements relative to
    some :class:`.DegreeOfFreedom` or driven :class:`.Parameter` in the model. The
    phases are optimised by calculating the DC response of the readouts. This
    :class:`.Action` changes the state of the model by varying the readout demodulation
    phases. If no arguments are given it will try to automatically optimise any lock
    element in the model that is using an RF readout with respect to the lock feedback
    parameter.

    Parameters
    ----------
    args
        Pairs of :class:`.DegreeOfFreedom` or :class:`.Parameter` and :class:`.ReadoutRF` elements, or pairs
        of their names. If none are provided `OptimiseRFReadoutPhaseDC` will
        automatically search for :class:`.Lock` elements which have :class:`.ReadoutRF`
        error signal and optimise them.
    d_dof : float, optional
        A small offset applied to the DOFs to compute the gradients of the error
        signals

    Examples
    --------
    Take a typicaly Pound-Drever-Hall lock of a cavity. Here is some KatScript
    to setup such a model:

    >>> import finesse
    >>> from finesse.analysis.actions import OptimiseRFReadoutPhaseDC
    >>>
    >>> model = finesse.Model()
    >>> model.parse('''
    >>> l l1
    >>> mod mod1 10M 0.1 mod_type=pm
    >>> readout_rf PD f=mod1.f phase=33 output_detectors=True optical_node=m1.p1.o
    >>> m m1 R=0.99 T=0.01
    >>> m m2 R=1 T=0
    >>> link(l1, mod1, m1, 1, m2)
    >>> lock cav_lock PD_I m2.phi 0.01 1e-3
    >>> ''')

    We have defined a `lock` above using the I quadrature RF demodulation and
    feeding back to the `m2` mirror position. We can optimise this demodulation
    phase by running. Here we manually provied which drives and readouts to use:

    >>> sol = model.run(OptimiseRFReadoutPhaseDC("m2.phi", 'PD_I'))
    >>> print(sol.phases)
    {'PD': 181.3535303754581}
    >>> print(model.PD.phase)
    181.3535303754581

    Alternatively, `PD_Q` could also be optimised for above. You can also just
    optimise all locks that are using RF readouts by providing no arguments:

    >>> sol = model.run(OptimiseRFReadoutPhaseDC())
    >>> print(sol.phases)
    {'PD': 181.3535303754581}

    To tell what was optimised, see the `sol.phases` dictionary.
    """

    def __init__(self, *args, d_dof=1e-10, name="optimise_demod_phases_dc"):
        super().__init__(name)
        self.args = args
        self.d_dof = d_dof

    def _do(self, state):
        readout_outputs = [
            state.sim.model.elements[name] for name in self.readout_output_names
        ]
        readouts = [state.sim.model.elements[name] for name in self.readout_names]
        drives = []

        Idws = get_readout_workspace(readouts, "I", state.sim.workspace_name_map)
        Qdws = get_readout_workspace(readouts, "Q", state.sim.workspace_name_map)
        drives = tuple(state.model.get(drive) for drive in self.drive_names)

        assert len(drives) == len(readouts) == len(readout_outputs)

        N = len(drives)
        sol = OptimiseRFReadoutPhaseDCSolution(self.name)
        sol.Ivals = np.zeros((N, 2), dtype=complex)
        sol.Qvals = np.zeros((N, 2), dtype=complex)

        # Here we compute the gradient of the error signals
        # with respect to some DOF change
        for i in range(N):
            drives[i].value -= self.d_dof
            state.sim.run_carrier()
            sol.Ivals[i, 0] = Idws[i].get_output()
            sol.Qvals[i, 0] = Qdws[i].get_output()
            drives[i].value += 2 * self.d_dof
            state.sim.run_carrier()
            sol.Ivals[i, 1] = Idws[i].get_output()
            sol.Qvals[i, 1] = Qdws[i].get_output()
            # reset value
            drives[i].value -= self.d_dof

        # Compute the gradients in both I and Q
        sol.I_gradients = (sol.Ivals[:, 1] - sol.Ivals[:, 0]) / 2e-6
        sol.Q_gradients = (sol.Qvals[:, 1] - sol.Qvals[:, 0]) / 2e-6

        # We can use the complex angle to compute how much to change the
        # demod phase by to optimise it
        sol.add_degrees = np.angle(sol.I_gradients + 1j * sol.Q_gradients, deg=True)
        sol.phases = {}
        sol.previous_phases = {}

        for i in range(N):
            param = readouts[i].phase
            sol.previous_phases[self.readout_names[i]] = float(param.value)
            if self.readout_output_names[i].endswith("_Q"):
                param.value += sol.add_degrees[i] + 90
            else:
                param.value += sol.add_degrees[i]
            sol.phases[self.readout_names[i]] = float(param.value)

        return sol

    def _setup(self, model):
        if len(self.args) == 0:
            rf_locks = tuple(
                lock
                for lock in model.locks
                if isinstance(lock.error_signal, ReadoutDetectorOutput)
                and isinstance(lock.error_signal.readout, ReadoutRF)
            )
            self.drive_names = list(lock.feedback.full_name for lock in rf_locks)
            self.readout_output_names = list(
                lock.error_signal.name for lock in rf_locks
            )
        else:
            self.drive_names = list(elements_to_name(self.args[::2]))
            self.readout_output_names = list(elements_to_name(self.args[1::2]))

        self.readout_names = []

        for i, name in enumerate(self.drive_names):
            obj = model.get(name)
            if isinstance(obj, DegreeOfFreedom):
                self.drive_names[i] = obj.DC.full_name
            elif isinstance(obj, Parameter):
                self.drive_names[i] = obj.full_name
            elif isinstance(obj, ParameterRef):
                self.drive_names[i] = obj.parameter.full_name
            else:
                raise FinesseException(
                    f"OptimiseRFReadoutPhaseDC: cannot drive {repr(obj)}"
                )

        for i, name in enumerate(self.readout_output_names):
            obj = model.get(name)
            if isinstance(obj, ReadoutDetectorOutput):
                self.readout_names.append(obj.readout.name)
            elif isinstance(obj, ReadoutRF):
                # first output defined, usually I
                default = obj.outputs.I.name
                deprecation_warning(
                    f"OptimiseRFReadoutPhaseDC should now be given a ReadoutDetectorOutput not ReadoutRF '{obj}'. Defaulting to output '{default}'",
                    "3.0",
                )
                self.readout_output_names[i] = default
                self.readout_names.append(obj.name)
            else:
                raise FinesseException(
                    f"OptimiseRFReadoutPhaseDC: {repr(obj)} should be a ReadoutRF or a ReadoutDetectorOutput"
                )

        if len(self.drive_names) != len(self.readout_output_names):
            raise FinesseException(
                "OptimiseRFReadoutPhaseDC: matching pairs of drives and readouts should be provided"
            )

    def _requests(self, model, memo, first=True):
        self._setup(model)
        memo["changing_parameters"].extend(_ for _ in self.drive_names)
        memo["changing_parameters"].extend(_ + ".phase" for _ in self.readout_names)
        return memo


class SensingMatrixSolution(BaseSolution):
    """Sensing matrix solution.

    The raw sensing matrix information can be accessed using the
    `SensingMatrixSolution.out` member. This is a complex-valued array with dimensions
    (DOFs, Readouts), which are accessible via `SensingMatrixSolution.dofs` and
    `SensingMatrixSolution.readouts`.

    A table can be printed using :meth:`.SensingMatrixSolution.display`.

    Polar plot can be generated using :meth:`.SensingMatrixSolution.plot`

    Printing :class:`.SensingMatrixSolution` will show an ASCII table of the data.
    """

    def display(
        self,
        dofs=None,
        readouts=None,
        tablefmt="pandas",
        numfmt="{:.2G}",
        highlight=None,
        highlight_color="#808080",
    ):
        """Displays a HTML table of the sensing matrix, optionally highlighting the
        largest absolute value for each readout or dof.

        Notes
        -----
        Only works when called from an IPython environment with the
        `display` method available. Pandas is required for highlighting.

        Parameters
        ----------
        dofs : iterable[str], optional
            Names of degrees of freedom to show, defaults to all if None
        readouts : iterable[str], optional
            Names of readouts to show, defaults to all if None
        tablefmt : str, optional
            Either 'pandas' for pandas formatting, or anything else to
            use `finesse.utilities.tables.Table`. Defaults to 'pandas' if available.
        numfmt : str or func or array, optional
            Either a function to format numbers or a formatting string. The
            function must return a string. Can also be an array with one option per
            row, column or cell. Defaults to '{:.2G}'.
        highlight : str or None, optional
            Either 'dof' to highlight the readout that gives the largest
            output for each dof, or 'readout' to highlight the dof for
            which each readout gives the largest output. Defaults to
            None (no highlighting).
        highlight_color : str, optional
            Color to highlight the maximum values with. Pandas is
            required for this to have an effect. Defaults to pale
            orange.
        """
        from IPython.display import display

        B, dofs, readouts = self.matrix_data(dofs, readouts)

        if tablefmt == "pandas":
            try:
                import pandas as pd
            except ModuleNotFoundError:
                tablefmt = "html"

        if tablefmt == "pandas":

            def highlight_max(data):
                return np.where(
                    abs(data) == abs(data).max(),
                    f"background-color: {highlight_color}",
                    "",
                )

            B = pd.DataFrame(B, index=dofs, columns=readouts)

            if highlight == "dof":
                style = B.style.apply(highlight_max, axis=1)
            elif highlight == "readout":
                style = B.style.apply(highlight_max, axis=0)
            elif highlight is None:
                style = B.style
            else:
                raise ValueError(
                    "Argument 'highlight' must be one of 'dof', 'readout' or None."
                )

            display(style.format(numfmt))
        else:
            return NumberTable(
                B,
                colnames=readouts,
                rownames=dofs,
                numfmt=numfmt,
            )

    def __str__(self):
        B, dofs, readouts = self.matrix_data()
        return str(NumberTable(B, colnames=readouts, rownames=dofs, numfmt="{:.2G}"))

    def matrix_data(self, dofs=None, readouts=None):
        """Generates a sensing matrix table.

        Parameters
        ----------
        dofs : iterable[str], optional
            Names of degrees of freedom to show, defaults to all if None
        readouts : iterable[str], optional
            Names of readouts to show, defaults to all if None

        Returns
        -------
        matrix : 2D numpy array, complex
        dofs : list of :class:`str`
        readouts: list of :class:`str`
        """
        dofs = dofs or self.dofs
        if readouts is not None:
            readouts = readouts
            readouts_rf = [rd for rd in self.readouts_rf if rd in readouts]
            readouts_dc = [rd for rd in self.readouts_dc if rd in readouts]
            try:
                readout_indices = [self.readouts.index(rd) for rd in readouts]
                A = self.out[:, readout_indices]
            except Exception:
                print(
                    "ValueError: Some readouts provided "
                    "are not present in the sensing matrix."
                )
                raise
        else:
            readouts = self.readouts
            readouts_rf = self.readouts_rf
            readouts_dc = self.readouts_dc
            A = self.out

        hdrs = []
        for rd in readouts:
            if rd in readouts_rf:
                hdrs.append(rd + "_I")
                hdrs.append(rd + "_Q")
            else:
                hdrs.append(rd + "_DC")
        Nd = len(dofs)
        Nr_rf = len(readouts_rf)
        Nr_dc = len(readouts_dc)
        B = np.zeros((Nd, ((2 * Nr_rf) + Nr_dc)))
        col_num = 0
        for ind, rd in enumerate(readouts):
            if rd in readouts_rf:
                B[:, col_num] = A[:, ind].real
                B[:, col_num + 1] = A[:, ind].imag
                col_num += 2
            else:
                B[:, col_num] = A[:, ind].real
                col_num += 1
        return B, dofs, hdrs

    def plot(
        self, Nrows, Ncols, figsize=(6, 5), *, dofs=None, readouts=None, r_lims=None
    ):
        import matplotlib.pyplot as plt

        dofs = np.atleast_1d(dofs or self.dofs)
        readouts = np.atleast_1d(readouts or self.readouts)

        fig, axs = plt.subplots(
            Nrows,
            Ncols,
            figsize=figsize,
            subplot_kw={"projection": "polar"},
            squeeze=False,
        )
        axs = axs.flatten()
        for idx in range(len(readouts)):
            dof_idxs = tuple(self.dofs.index(_) for _ in dofs)
            _ax = axs[idx]
            A = self.out[dof_idxs, idx]

            _ax.set_theta_zero_location("E")
            if r_lims is None or (r_lims is not None and r_lims[idx] is None):
                r_lim = (np.log10(np.abs(A)).min() - 1, np.log10(np.abs(A)).max())
            else:
                r_lim = np.log10(r_lims[idx])

            _ax.set_ylim(r_lim[0], r_lim[1] + 1)
            _ax.set_yticklabels([])

            theta = np.angle(A)
            r = np.log10(np.abs(A))
            _ax.plot(
                (theta, theta),
                (r_lim[0] * np.ones_like(r), r),
                marker="D",
                markersize=5,
            )
            _ax.set_title(self.readouts[idx])
            _ax.set_ylim(r_lim[0], r_lim[1] + 1)

        _ax.legend(self.dofs, loc="best", bbox_to_anchor=(0.5, -0.3), fontsize=8)
        plt.tight_layout(pad=1.2)
        return fig, axs


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class SensingMatrixDC(Action):
    """Computes the sensing matrix elements for various degrees of freedom and readouts
    that should be present in the model. The solution object for this action then
    contains all the information on the sensing matrix. This can be plotted in polar
    coordinates, displayed in a table, or directly accessed.

    The sensing gain is computed by calculating the gradient of each readout
    signal, which means it is a DC measurement. This will not include any
    suspension or radiation pressure effects.

    This action does not modify the states model.

    Parameters
    ----------
    dofs : iterable[str]
        String names of degrees of freedom
    readouts : iterable[str]
        String names of readouts
    d_dof : float, optional
        Small step used to compute derivative
    """

    def __init__(self, dofs, readouts, d_dof=1e-9, name="sensing_matrix_dc"):
        super().__init__(name)
        # only store string names
        self.dofs = elements_to_name(dofs)
        self.readouts = elements_to_name(readouts)
        self.d_dof = d_dof

    def _do(self, state):
        self.readouts_rf = []
        self.readouts_dc = []
        Idws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_I", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        Qdws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_Q", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        DCws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == rd + "_DC", state.sim.readout_workspaces
                ),
                None,
            )
            for rd in self.readouts
        )
        dcs = tuple(state.model.get(f"{dof}.DC") for dof in self.dofs)
        Nd = len(self.dofs)
        Nr = len(self.readouts)

        sol = SensingMatrixSolution(self.name)
        sol.dofs = self.dofs
        sol.readouts = self.readouts
        sol.readouts_rf = []
        sol.readouts_dc = []
        sol.vals = np.zeros((Nd, Nr, 2), dtype=complex)
        sol.out = np.zeros((Nd, Nr), dtype=complex)
        # Here we compute the gradient of the error signals
        # with respect to some DOF change
        for i in range(Nd):
            dcs[i].value -= self.d_dof
            state.sim.run_carrier()
            for j in range(Nr):
                if Idws[j] is not None:
                    sol.vals[i, j, 0] += Idws[j].get_output()
                    sol.vals[i, j, 0] += 1j * Qdws[j].get_output()
                    if i == 0:
                        sol.readouts_rf.append(self.readouts[j])
                else:
                    sol.vals[i, j, 0] += DCws[j].get_output()
                    if i == 0:
                        sol.readouts_dc.append(self.readouts[j])
            dcs[i].value += 2 * self.d_dof
            state.sim.run_carrier()
            for j in range(Nr):
                if Idws[j] is not None:
                    sol.vals[i, j, 1] += Idws[j].get_output()
                    sol.vals[i, j, 1] += 1j * Qdws[j].get_output()
                else:
                    sol.vals[i, j, 1] += DCws[j].get_output()
            # reset value
            dcs[i].value -= self.d_dof

        # Compute the gradients
        sol.out = (sol.vals[:, :, 1] - sol.vals[:, :, 0]) / (2 * self.d_dof)
        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].extend((f"{_}.DC" for _ in self.dofs))
        return memo


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class SensingMatrixAC(Action):
    """Computes the sensing matrix elements for various degrees of freedom and readouts
    that should be present in the model. The solution object for this action then
    contains all the information on the sensing matrix. This can be plotted in polar
    coordinates, displayed in a table, or directly accessed.

    The sensing gain is computed by calculating the gradient of each readout
    signal, which means it is a DC measurement. This will not include any
    suspension or radiation pressure effects.

    This action does not modify the states model.

    Parameters
    ----------
    dofs : iterable[str]
        String names of degrees of freedom
    readouts : iterable[str]
        String names of readouts
    f : float
        Frequency to measure sensing matrix at
    """

    def __init__(self, dofs, readouts, f=1e-3, name="sensing_matrix_ac"):
        super().__init__(name)
        # only store string names
        self.dofs = elements_to_name(dofs)
        self.readouts = elements_to_name(readouts)
        self.f = f

        self.nodes = []
        self.nodes.extend([readout + ".I" for readout in self.readouts])
        self.nodes.extend([readout + ".Q" for readout in self.readouts])

    def _do(self, state):
        sol = SensingMatrixSolution(self.name)
        sol.dofs = self.dofs
        sol.readouts = self.readouts

        sol.freqresp = FrequencyResponse((self.f,), self.dofs, self.nodes)._do(state)

        sol.out = np.zeros((len(self.dofs), len(self.readouts)), dtype=np.complex128)
        for i, dof in enumerate(self.dofs):
            for j, readout in enumerate(self.readouts):
                sol.out[i, j] = np.real(sol.freqresp[dof, readout + ".I"])
                sol.out[i, j] += 1j * np.real(sol.freqresp[dof, readout + ".Q"])

        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append("fsig.f")
        memo["input_nodes"].extend((dof, ("input",)) for dof in self.dofs)
        memo["output_nodes"].extend((node, ("output",)) for node in self.nodes)
        return memo


class CheckLinearitySolution(BaseSolution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.results = None
        self.lock_names = ()


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class CheckLinearity(Action):
    """An action that shows the relationships between all DOFs and all error signals, to
    check whether they are related linearly. Plotted for DOFs starting at their initial
    values and up until their initial values + 2*gain*intial error signal.

    Parameters
    ----------
    \\*locks : list, optional
        A list of locks to use in each :class:`finesse.analysis.actions.locks.RunLocks`
        step. Acts like \\*locks parameter in
        :class:`finesse.analysis.actions.locks.RunLocks`: if not provided, all locks in
        model are used.

    num_points : int
        Number of points to plot in the DOF range.

    plot_results : boolean
        Whether or not to plot results (requires
        matplotlib)

    xlim : list or None
        Defines (half of) the range of DOF values
        over which to plot the error signals. If
        not specified, gains are used to find a
        useful range of DOF values to plot over.

    name : str
        Name of the action.
    """

    def __init__(
        self, *locks, num_points=10, plot_results=True, xlim=None, name="run locks"
    ):
        super().__init__(name)
        self.locks = elements_to_name(locks)
        # Round up to the nearest odd integer, so that the plot always
        # includes the current points.
        self.num_points = num_points + 1 if num_points % 2 == 0 else num_points
        self.xlim = xlim
        self.plot_results = plot_results

    def _do(self, state):
        if state.sim is None:
            raise Exception("Simulation has not been built")
        if not isinstance(state.sim, SparseMatrixSimulation):
            raise NotImplementedError()

        if len(self.locks) == 0:
            locks = tuple(lck for lck in state.model.locks if lck.enabled)
        else:
            locks = tuple(
                state.model.elements[name]
                for name in self.locks
                if not state.model.elements[name].enabled
            )

        if self.xlim is not None:
            if len(self.xlim) != len(locks):
                raise Exception("Number of locks and xlim not equal.")
            # else:
            # xlim = self.xlim # Not used

        out_wss = OrderedSet(  # workspaces can be in both lists
            (*state.sim.readout_workspaces, *state.sim.detector_workspaces)
        )

        dws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == lock.error_signal.name,
                    out_wss,
                ),
                None,
            )
            for lock in locks
        )
        sol = CheckLinearitySolution(self.name)
        N = len(locks)
        # Store initial parameters in case of failure so we can reset the model
        initial_parameters = tuple(float(lock.feedback) for lock in locks)
        initial_errors = tuple(
            float(dw.get_output() - locks[dws.index(dw)].offset) for dw in dws
        )
        sol.results = np.zeros((N, N, 2, self.num_points))
        sol.lock_names = tuple(lock.name for lock in locks)

        err_sigs = [lck.error_signal for lck in locks]
        err_sig_names = [sig.name for sig in err_sigs]
        readout_names = [sig.readout.name for sig in err_sigs]
        lock_dof_names = [deref(lck.feedback).owner.name for lck in locks]

        sensing_matrix = state.apply(SensingMatrixDC(lock_dof_names, readout_names))
        gain_matrix = np.zeros((N, N))
        for dof_idx in range(N):
            for rd_idx in range(N):
                err_sig = err_sig_names[rd_idx]
                val = sensing_matrix.out[dof_idx, rd_idx]
                if "_Q" in err_sig:
                    gain = val.imag
                else:
                    gain = val.real
                gain_matrix[rd_idx, dof_idx] = gain

        # Index i runs over error signals
        for i in range(N):
            # Index j runs over DOFs
            initial_error = initial_errors[i]
            for j in range(N):
                initial_param = initial_parameters[j]

                if self.xlim is not None:
                    dof_list = np.linspace(
                        initial_param - self.xlim[j],
                        initial_param + self.xlim[j],
                        self.num_points,
                    )
                elif gain_matrix[i, j] == 0:
                    dof_list = np.linspace(
                        initial_param - 1, initial_param + 1, self.num_points
                    )
                else:
                    lock_gain = -1 / gain_matrix[i, j]
                    dof_list = np.linspace(
                        initial_param - 0 * lock_gain * initial_error,
                        initial_param + 2 * lock_gain * initial_error,
                        self.num_points,
                    )

                rel_dof_list = dof_list - initial_param
                sol.results[i, j, 0] = rel_dof_list
                for idx, dof_val in enumerate(dof_list):
                    deref(locks[j].feedback).value = dof_val
                    state.sim.run_carrier()
                    new_error = dws[i].get_output() - locks[i].offset
                    sol.results[i, j, 1, idx] = new_error
                deref(locks[j].feedback).value = initial_param

        if self.plot_results:
            import matplotlib.pyplot as plt

            plt.rcParams["figure.figsize"] = [1.5 * N, 1.5 * N]
            if N > 1:
                fig, axs = plt.subplots(N, N)
                for i in range(N):
                    for j in range(N):
                        axs[i][j].plot(
                            sol.results[i, j, 0, 0:], sol.results[i, j, 1, 0:], zorder=0
                        )
                        axs[i][j].ticklabel_format(
                            axis="y", style="sci", scilimits=(0, 0)
                        )
                for ax, name in zip(axs[-1], lock_dof_names):
                    ax.set_xlabel(name, labelpad=10, fontsize=14)
                for ax, name in zip(axs[:, 0], err_sig_names):
                    ax.set_ylabel(name, labelpad=10, fontsize=14)
                plt.tight_layout()
                plt.subplots_adjust(wspace=0.6, hspace=0.6)
            elif N == 1:
                plt.plot(sol.results[0, 0, 0, 0:], sol.results[0, 0, 1, 0:])
                plt.xlabel(lock_dof_names[0], fontsize=14)
                plt.ylabel(err_sig_names[0], fontsize=14)
            else:
                print("No existing locks to display.")

            plt.show()
        return sol

    def _requests(self, model, memo, first=True):
        if len(self.locks) == 0:
            # If none given lock everything
            for lock in model.locks:
                memo["changing_parameters"].append(deref(lock.feedback).full_name)
                rd_name = lock.error_signal.name
                if "_DC" not in rd_name:
                    memo["changing_parameters"].append(
                        lock.error_signal.readout.name + ".phase"
                    )
        else:
            for name in self.locks:
                if name not in model.elements:
                    raise Exception(f"Model {model} does not have a lock called {name}")
                memo["changing_parameters"].append(
                    model.elements[name].feedback.parameter.full_name
                )
                rd_name = model.elements[name].error_signal.name
                if "_DC" not in rd_name:
                    memo["changing_parameters"].append(
                        model.elements[name].error_signal.readout.name + ".phase"
                    )


class GetErrorSignalsSolution(BaseSolution):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.results = None
        self.lock_names = ()


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class GetErrorSignals(Action):
    """An action that quickly calculates the current  error signals for all or a subset
    of locks in a model.

    Parameters
    ----------
    \\*locks : list, optional
        A list of lock names to compute the error signals for.
        If not provided, all locks in model are used.

    name : str
        Name of the action.
    """

    def __init__(self, *locks, name="get error signals"):
        super().__init__(name)
        self.locks = elements_to_name(locks)

    def _do(self, state):
        if state.sim is None:
            raise Exception("Simulation has not been built")
        if not isinstance(state.sim, SparseMatrixSimulation):
            raise NotImplementedError()

        if len(self.locks) == 0:
            locks = state.model.locks
        else:
            locks = tuple(state.model.elements[lock] for lock in self.locks)

        out_wss = OrderedSet(  # workspaces can be in both lists so combine them
            (*state.sim.readout_workspaces, *state.sim.detector_workspaces)
        )

        dws = tuple(
            next(
                filter(
                    lambda x: x.oinfo.name == lock.error_signal.name,
                    out_wss,
                ),
                None,
            )
            for lock in locks
        )

        state.sim.run_carrier()
        N = len(locks)
        sol = GetErrorSignalsSolution(self.name)
        sol.results = np.zeros(N)
        sol.lock_names = tuple(lock.name for lock in locks)
        for i in range(N):
            res = dws[i].get_output() - locks[i].offset
            sol.results[i] = res

        return sol

    def _requests(self, model, memo, first=True):
        pass



import pytest
import numpy as np


@pytest.fixture
def kat(model):
    model.parse(
        """
        l l1 P=1 phase=2.19571

        readout_dc REFL l1.p1.o

        pd1 REFL_DC1 l1.p1.o fsig
        ad REFL_DC2 REFL.DC.o fsig

        fsig(1)
        """
    )
    return model


def test_pwr(kat):
    kat.parse("sgen DIFF l1.pwr.i phase=10")
    sol = kat.run()
    assert np.allclose(sol["REFL_DC1"], sol["REFL_DC2"])
    assert np.allclose(abs(sol["REFL_DC1"]), 1)
    assert np.allclose(np.rad2deg(np.angle(sol["REFL_DC1"])), 10)


def test_amp(kat):
    kat.parse("sgen DIFF l1.amp.i phase=10")
    sol = kat.run()
    assert np.allclose(sol["REFL_DC1"], sol["REFL_DC2"])
    assert np.allclose(abs(sol["REFL_DC1"]), 2)
    assert np.allclose(np.rad2deg(np.angle(sol["REFL_DC1"])), 10)


def test_frq(kat):
    kat.parse("sgen DIFF l1.frq.i phase=10")
    sol = kat.run()
    assert np.allclose(sol["REFL_DC1"], sol["REFL_DC2"])
    assert np.allclose(abs(sol["REFL_DC1"]), 0)


def test_phs(kat):
    kat.parse("sgen DIFF l1.phs.i phase=10")
    sol = kat.run()
    assert np.allclose(sol["REFL_DC1"], sol["REFL_DC2"])
    assert np.allclose(abs(sol["REFL_DC1"]), 0)



def clip_with_tolerance(
    number: float, _min: float, _max: float, tol: float = 1e-12
) -> float:
    """Clips a number to the range [_min, _max] if it is in [_min - tol, _max + tol],
    otherwise just returns the number.

    Useful for correcting numerical precision errors after calculating physical
    quantities.

    Parameters
    ----------
    number : float
        Number to clip
    _min : float
        lower end of clipping range
    _max : float
        higher end of clipping range
    tol : float, optional
        Absolute tolerance, by default 1e-12

    Returns
    -------
    float
        The clipped number
    """
    if number < _min and (number + tol) > _min:
        return _min
    elif number > _max and (number - tol) < _max:
        return _max
    else:
        return number



from finesse import Model


def test_model_run_no_detectors():
    """Ensure model runs even if no detectors present."""
    model = Model()
    model.parse(
        """
        l L0 P=1
        s s1 L0.p1 ETM.p1
        m ETM T=1e-5 L=1e-6 Rc=10
        """
    )
    model.run()




import logging

cimport cython

from finesse.cymath.cmatrix import SubCCSView
from finesse.cymath.cmatrix cimport SubCCSView
from ..components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from ..components.workspace import Connections
from ..cymath cimport complex_t
from ..components import Connector, NodeType, NodeDirection
from ..parameter import float_parameter
from ..components.general import LocalDegreeOfFreedom
from finesse.exceptions import FinesseException
from ..utilities import zpk_fresp
from finesse.components.general import MechanicalConnector

from libc.stdlib cimport free, calloc

cimport numpy as np
import numpy as np

import types
import itertools


LOGGER = logging.getLogger(__name__)

cdef extern from "constants.h":
    long double PI


cdef class MIMOTFWorkspace(ConnectorWorkspace):
    """Workspace that contains MIMO transfer functions stored
    in a numerator/denominator basis.
    """

    cdef:
        double[::1] denom
        Py_ssize_t N_num_allocd
        Py_ssize_t total_numerators
        double** numerators
        int* numerator_sizes
        complex_t curr_denom
        complex_t s

    def __cinit__(self, owner, sim, bint refill, unsigned int N_numerators):
        self.N_num_allocd = N_numerators
        self.numerators = <double**>calloc(N_numerators, sizeof(double*))
        if not self.numerators:
            raise MemoryError()
        self.numerator_sizes = <int*>calloc(N_numerators, sizeof(int))
        if not self.numerator_sizes:
            raise MemoryError()
        self.total_numerators = 0
        self.curr_denom = 0

    def __init__(self, owner, sim, bint refill, unsigned int N_numerators):
        super().__init__(
            owner,
            sim,
            Connections(),
            Connections()
        )
        self.signal.add_fill_function(mimo_fill, refill)

    def __dealloc__(self):
        if self.numerators:
            free(self.numerators)
        if self.numerator_sizes:
            free(self.numerator_sizes)

    @property
    def num_numerators(self):
        return self.total_numerators

    def set_denominator(self, double[::1] denom):
        self.denom = denom

    def add_numerator(self, double[::1] num):
        if self.total_numerators == self.N_num_allocd:
            raise Exception("Added more numerators than were allocated for.")
        self.numerators[self.total_numerators] = &num[0]
        self.numerator_sizes[self.total_numerators] = len(num)
        self.total_numerators += 1

    cpdef void set_s(self, complex_t s) noexcept:
        self.s = s
        self.curr_denom = eval_tf_term(s, &self.denom[0], len(self.denom))

    cpdef complex_t H(self, int numerator_idx) noexcept:
        if not (0 <= numerator_idx < self.total_numerators):
            raise Exception("Unexpected index")

        return eval_tf(
            self.s,
            self.numerators[numerator_idx],
            self.numerator_sizes[numerator_idx],
            self.curr_denom
        )


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef inline complex_t eval_tf_term(complex_t s, const double* coeffs, int N) noexcept:
    cdef:
        int i
        complex res = 0
    for i in range(N):
        res = res * s + coeffs[i]

    return res


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef inline complex_t eval_tf(complex_t s, const double* num, int N, complex_t den) noexcept:
    return eval_tf_term(s, num, N)/den


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cpdef eval_tf_vec(const complex_t[::1] s, const double[::1] num, const double[::1] den, complex_t[::1] out) :
    cdef:
        int i
        int N = len(s)
        int Nn = len(num)
        int Nd = len(den)

    if len(out) != len(s):
        raise Exception("Length of `s` differs from output `out`")

    for i in range(N):
        out[i] = eval_tf_term(s[i], &num[0], Nn)/eval_tf_term(s[i], &den[0], Nd)

    return 0

mimo_fill = FillFuncWrapper.make_from_ptr(c_mimo_fill)
@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef c_mimo_fill(ConnectorWorkspace cws) :
    cdef MIMOTFWorkspace ws = <MIMOTFWorkspace>cws
    cdef complex_t s = 0
    cdef tuple key
    s.imag = 2 * PI * ws.sim.model_settings.fsig
    # Sets the complex s value for this step and precomputes the denominator
    ws.set_s(s)
    for i in range(ws.total_numerators):
        key = (ws.owner_id, i, 0, 0)
        if key in ws.sim.signal._submatrices:
            (<SubCCSView>ws.sim.signal._submatrices[key]).fill_negative_za(ws.H(i))


class FreeMassWorkspace(ConnectorWorkspace):
    pass


@float_parameter("mass", "Mass", units="kg")
@float_parameter("I_pitch", "Moment of inertia (pitch)", units="kg·m^2")
@float_parameter("I_yaw", "Moment of inertia (yaw)", units="kg·m^2")
class FreeMass(MechanicalConnector):
    """Simple free mass suspension of an object.

    The object being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.
    """

    def __init__(self, name, connected_to, mass=np.inf, I_yaw=np.inf, I_pitch=np.inf):
        super().__init__(name, connected_to)
        self.mass = mass
        self.I_yaw = I_yaw
        self.I_pitch = I_pitch

        # We just have direct coupling between DOF, no cross-couplings
        self._register_node_coupling(
            "F_to_Z", self.mech.F_z, self.mech.z,
            enabled_check=lambda: float(self.mass) < np.inf or self.mass.is_changing
        )
        self._register_node_coupling(
            "F_to_YAW", self.mech.F_yaw, self.mech.yaw,
            enabled_check=lambda: float(self.I_yaw) < np.inf or self.I_yaw.is_changing
        )
        self._register_node_coupling(
            "F_to_PITCH", self.mech.F_pitch, self.mech.pitch,
            enabled_check=lambda: float(self.I_pitch) < np.inf or self.I_pitch.is_changing
        )
        # Define typical degrees of freedom for this component
        import types
        self.dofs = types.SimpleNamespace()
        self.dofs.z = LocalDegreeOfFreedom(f"{self.name}.dofs.z" ,None, self.mech.z, 1)
        self.dofs.F_z = LocalDegreeOfFreedom(f"{self.name}.dofs.F_z", None, self.mech.F_z, 1, AC_OUT=self.mech.z)
        self.dofs.yaw = LocalDegreeOfFreedom(f"{self.name}.dofs.yaw", None, self.mech.yaw, 1)
        self.dofs.F_yaw = LocalDegreeOfFreedom(f"{self.name}.dofs.F_yaw", None, self.mech.F_yaw, 1, AC_OUT=self.mech.yaw)
        self.dofs.pitch = LocalDegreeOfFreedom(f"{self.name}.dofs.pitch", None, self.mech.pitch, 1)
        self.dofs.F_pitch = LocalDegreeOfFreedom(f"{self.name}.dofs.F_pitch", None, self.mech.F_pitch, 1, AC_OUT=self.mech.pitch)


    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing or any(p.is_changing for p in self.parameters)
            ws = FreeMassWorkspace(self, sim)
            ws.signal.add_fill_function(self.fill, refill)
            return ws
        else:
            return None

    def fill(self, ws):
        f = ws.sim.model_settings.fsig
        if ws.signal.connections.F_to_Z_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_Z_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.mass * (2*PI*f)**2)

        if ws.signal.connections.F_to_YAW_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_YAW_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.I_yaw * (2*PI*f)**2)

        if ws.signal.connections.F_to_PITCH_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_PITCH_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.I_pitch * (2*PI*f)**2)


class PendulumMassWorkspace(ConnectorWorkspace):
    pass


@float_parameter("mass", "Mass", units="kg")
@float_parameter("Qz", "Qz", units="")
@float_parameter("fz", "fz", units="Hz")
@float_parameter("I_pitch", "Moment of inertia (pitch)", units="kg·m^2")
@float_parameter("Qyaw", "Qyaw", units="")
@float_parameter("fyaw", "fyaw", units="Hz")
@float_parameter("I_yaw", "Moment of inertia (yaw)", units="kg·m^2")
@float_parameter("Qpitch", "Qpitch", units="")
@float_parameter("fpitch", "fpitch", units="Hz")
class Pendulum(MechanicalConnector):
    """Simple pendulum suspension of an object.

    The object being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.
    """

    def __init__(self, name, connected_to, mass=np.inf, Qz=1000, fz=1, I_yaw=np.inf, Qyaw=1000, fyaw=1, I_pitch=np.inf, Qpitch=1000, fpitch=1):
        super().__init__(name, connected_to)
        self.mass = mass
        self.Qz = Qz
        self.fz = fz
        self.Qyaw = Qyaw
        self.fyaw = fyaw
        self.Qpitch = Qpitch
        self.fpitch = fpitch
        self.I_yaw = I_yaw
        self.I_pitch = I_pitch

        # We just have direct coupling between DOF, no cross-couplings
        self._register_node_coupling(
            "F_to_Z", self.mech.F_z, self.mech.z,
            enabled_check=lambda: float(self.mass) < np.inf or self.mass.is_changing
        )
        self._register_node_coupling(
            "F_to_YAW", self.mech.F_yaw, self.mech.yaw,
            enabled_check=lambda: float(self.I_yaw) < np.inf or self.I_yaw.is_changing
        )
        self._register_node_coupling(
            "F_to_PITCH", self.mech.F_pitch, self.mech.pitch,
            enabled_check=lambda: float(self.I_pitch) < np.inf or self.I_pitch.is_changing
        )

        self.dofs = types.SimpleNamespace()
        self.dofs.z = LocalDegreeOfFreedom(
            f"{self.name}.dofs.z", None, self.mech.z, 1
        )
        self.dofs.F_z = LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_z", None, self.mech.F_z, 1, AC_OUT=self.mech.z
        )
        self.dofs.pitch = LocalDegreeOfFreedom(
            f"{self.name}.dofs.pitch", None, self.mech.pitch, 1
        )
        self.dofs.F_pitch= LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_pitch", None, self.mech.F_pitch, 1, AC_OUT=self.mech.pitch
        )
        self.dofs.yaw = LocalDegreeOfFreedom(
            f"{self.name}.dofs.yaw", None, self.mech.yaw, 1
        )
        self.dofs.F_yaw= LocalDegreeOfFreedom(
            f"{self.name}.dofs.F_yaw", None, self.mech.F_yaw, 1, AC_OUT=self.mech.yaw
        )

    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing or any(p.is_changing for p in self.parameters)
            ws = PendulumMassWorkspace(self, sim)
            ws.signal.add_fill_function(self.fill, refill)
            return ws
        else:
            return None

    def fill(self, ws):
        cdef:
            complex_t s = 2j* PI * ws.sim.model_settings.fsig
            double omega0

        if ws.signal.connections.F_to_Z_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_Z_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fz
                mat[:] = 1 / ws.values.mass * 1/(s**2  + s * omega0/ws.values.Qz + omega0**2)

        if ws.signal.connections.F_to_YAW_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_YAW_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fyaw
                mat[:] = 1 / ws.values.I_yaw * 1/(s**2  + s * omega0/ws.values.Qyaw + omega0**2)

        if ws.signal.connections.F_to_PITCH_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_PITCH_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fpitch
                mat[:] = 1 / ws.values.I_pitch * 1/(s**2  + s * omega0/ws.values.Qpitch + omega0**2)


class SuspensionZPKWorkspace(ConnectorWorkspace):
    pass


class SuspensionZPK(MechanicalConnector):
    """A suspension that models multiple poles and zeros for the z, yaw, or pitch motion of an optic.
    The user must ensure that minus signs are correct for this transfer function as well as defining
    complex conjugae pairs for physically correct behaviour.

    ZPK terms are in units of radians/s.

    Parameters
    ----------
    name : str
        Element name
    connected_to : Element or mechanical port
        Mechanical port or element to attach this suspension to
    zpk_plant : dict
        Dictionary of {(output, input):(z,p,k)}

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> import finesse
    >>> from finesse.components.mechanical import SuspensionZPK
    >>>
    >>> model = finesse.Model()
    >>> model.fsig.f = 1
    >>> model.parse("m m1 R=1 T=0")
    >>> zpk_plant = {}
    >>> # F_z to z (longitudinal force to displacement)
    >>> zpk_plant['z', 'F_z'] = ([], [10], 1)
    >>> model.add(SuspensionZPK('sus', model.m1.mech, zpk_plant))
    >>> out = model.run("frequency_response(geomspace(1m, 100, 100), m1.mech.F_z, m1.mech.z)")
    >>> plt.loglog(out.f, abs(out['m1.mech.F_z', 'm1.mech.z']))
    """

    def __init__(self, name, connected_to, zpk_plant):
        super().__init__(name, connected_to)
        self.zpk_plant = zpk_plant
        self.zpks = []

        for (output, input), zpk in zpk_plant.items():
            self._register_node_coupling(
                f"{input}_to_{output}",
                getattr(self.mech, input),
                getattr(self.mech, output),
            )
            self.zpks.append(zpk) # store ordered zpks

    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing
            ws = SuspensionZPKWorkspace(self, sim)
            ws.signal.add_fill_function(self.fill, refill)
            ws.zpks = [(np.array(z), np.array(p), float(k)) for z,p,k in self.zpks]
            return ws
        else:
            return None

    def fill(self, ws):
        w = 2 * np.pi * ws.sim.model_settings.fsig

        for i, (z, p, k) in enumerate(ws.zpks):
            H = zpk_fresp(z, p, k, w)
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, i, 0, 0,
            ) as mat:
                mat[:] = H


class SuspensionTFPlant(Connector):
    """A customised suspension element that accepts arbitrary input and output
    nodes and a 2D array of `control.TransferFunction` objects that define
    a transfer function between each of them. See the Python Control package for
    more details on these.

    - Inputs and outputs must be named in the format `port.name`
    - Inputs must be unique to outputs
    - All plant transfer functions must have the same denominators

    The mechanical port of other optical elements can be connected to this suspension
    by using the `connections_to` dictionary input. This takes the name of a port
    defined in this suspension and maps it to the mechanical port of another element.

    The plant, inputs, and outputs cannot be changed once initialised. There is no
    KatScript interface for this component.

    Parameters
    ----------
    name : str
        Name of element
    inputs : array_like, list
        Sequence of inputs for this plant, should be strings of the format `port.node`
    outputs : array_like, list
        Sequence of outputs for this plant, should be strings of the format `port.node`
    plant : array_like, list[list]
        A 2D array of `control.TransferFunction` that describe the transfer function between
        each input to every output. Shape should be `[len(outputs), len(inputs)]`. Elements
        can also be `None` which means no coupling between each transfer function.
    connections_to : dict[str: SignalNode], optional
        Dict of port names at this suspension component and a mechanical port of some
        other element. Names of nodes at in each port should be identical. Keys of the
        dict should be in the format `port.node`. Values should be mechanical nodes of
        other elements.
    """

    def __init__(self, name, inputs, outputs, plant, connections_to=None):
        super().__init__(name)
        self.__inputs = np.atleast_1d(inputs).tolist()
        self.__outputs = np.atleast_1d(outputs).tolist()
        self.__plant = np.atleast_2d(plant).tolist()
        self.__connections_to = {} if connections_to is None else connections_to

        Ni = len(inputs)
        No = len(outputs)

        if len(inputs) != len(set(inputs)):
            raise ValueError("Inputs should all be unique")
        if len(outputs) != len(set(outputs)):
            raise ValueError("Outputs should all be unique")

        if not all(len(_.split(".")) == 2 for _ in inputs):
            raise ValueError("Inputs should all be in the format `port.node`")
        if not all(len(_.split(".")) == 2 for _ in outputs):
            raise ValueError("Outputs should all be in the format `port.node`")
        if np.array(plant).shape != (No, Ni):
            raise ValueError("Mechanical plant should be an N_outputs x N_inputs matrix of transfer functions")

        self.__input_indices = {n:i for i,n in enumerate(inputs)}
        self.__output_indices = {n:i for i,n in enumerate(outputs)}

        split_inputs = tuple(_.split(".") for _ in inputs)
        split_outputs = tuple(_.split(".") for _ in outputs)
        # Get the unique ports
        ports = list(set(port for port, _ in itertools.chain.from_iterable((split_inputs, split_outputs))))
        # replace with actual port
        ports = {port: self._add_port(port, NodeType.MECHANICAL) for port in ports}
        nodes = {}

        for port, node in itertools.chain.from_iterable((split_inputs, split_outputs)):
            p = ports[port]
            port_name = f"{port}.{node}"

            if port in connections_to:

                other_port = connections_to[port]
                if other_port.type != NodeType.MECHANICAL:
                    raise FinesseException(f"{other_port!r} is not a mechanical port")
                try:
                    other_node = other_port.node(node)
                except KeyError as ex:
                    raise FinesseException(f"Problem mapping `{port}.{node}` to {port!r}. It does not have a node called `{node}`.")
                if port_name in self.__input_indices:
                    self.__input_indices[other_node.port_name] = self.__input_indices[port_name]
                else:
                    self.__output_indices[other_node.port_name] = self.__output_indices[port_name]
                nodes[port_name] = p._add_node(node, None, other_node)
            else:
                nodes[port_name] = p._add_node(node, NodeDirection.BIDIRECTIONAL)

        for i, I in enumerate(inputs):
            for o, O in enumerate(outputs):
                # If there is a connection between the nodes...
                if plant[o][i] is not None and plant[o][i].dcgain():
                    self._register_node_coupling(
                        f"{I}__{O}".replace(".", "_"), nodes[I], nodes[O]
                    )

    def bode(self, f, input_node, output_node, **kwargs):
        """Make a bode plot for a particular node coupling for this suspension.
        See `finesse.plotting.bode` for the actual plotting method.
        """

        from finesse.plotting.plot import bode
        try:
            i = self.__input_indices[input_node]
        except KeyError as ex:
            i = self.__input_indices[input_node.port_name]
        try:
            o = self.__output_indices[output_node]
        except KeyError as ex:
            o = self.__output_indices[output_node.port_name]

        if self.plant[o][i] is None:
            raise FinesseException(f"No connection between {input_node!r} and  {output_node!r}")
        Y = np.squeeze(self.plant[o][i].horner(2j*np.pi*f))
        return bode(f, Y, **kwargs)

    @ property
    def connections_to(self):
        return self.__connections_to.copy()

    @property
    def inputs(self):
        return self.__inputs

    @property
    def outputs(self):
        return self.__outputs

    @property
    def plant(self):
        return self.__plant

    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing  # Need to recompute H(f)
            N = len(self._registered_connections)
            ws = MIMOTFWorkspace(self, sim, refill, N)
            den_set = False
            # All TFs should have same denominators
            for tf in itertools.chain.from_iterable(self.plant):
                if tf is not None:
                    den_set = True
                    ws.set_denominator(tf.den[0][0])
                    break
            if not den_set :
                raise FinesseException("No denominators for setting up MIMO workspace")
            # Setup the TFs for filling
            for j, (i, o) in enumerate(self._registered_connections.values()):
                i = self.nodes[i]
                o = self.nodes[o]
                idx = self.__input_indices[i.port_name]
                odx = self.__output_indices[o.port_name]
                # Check if these nodes are actually being modelled or not, might have
                # been removed if not needed.
                if i.full_name in sim.signal.nodes and o.full_name in sim.signal.nodes:
                    if self.plant[odx][idx] is not None:
                        ws.add_numerator(self.plant[odx][idx].num[0][0])
            return ws
        else:
            return None



from finesse.cymath.gaussbeam cimport beam_param


cdef struct NodeBeamParam:
    beam_param qx
    beam_param qy
    bint is_fixed # Whether the q values will be changing during a simulation


cdef class PhaseConfig:
    cdef public:
        bint zero_k00 # should phase of k0000 coefficients be zeroed
        bint zero_tem00_gouy # should Gouy phase of TEM00 be zeroed
        bint v2_transmission_phase


cdef class ModelSettings:
    cdef public:
        double fsig
        double EPSILON0_C
        double UNIT_VACUUM
        double x_scale
        bint is_modal
        int[:, ::1] homs_view
        PhaseConfig phase_config
        double lambda0, f0, k0
        int num_HOMs
        int max_n # Maximum mode index in tangential plane
        int max_m # Maximum mode index in sagittal plane


# Contains non-physical configuration data, as low-level values, to be
# used during a simulation
cdef struct SimConfigData:
    # Number of threads to use for prange loops whose
    # size is proportional to the number of modes
    int nthreads_homs



"""Actions to compute noise projections and budgets."""

import logging
from collections import defaultdict

import numpy as np

from ...detectors.compute.quantum import QShot0Workspace, QShotNWorkspace
from ...env import warn
from ...solutions import BaseSolution
from .base import Action
from .lti import FrequencyResponse

LOGGER = logging.getLogger(__name__)


class NoiseProjectionSolution(BaseSolution):
    def plot(self, output_node=None, lower=0.1, upper=3, *, ax=None, **kwargs):
        import matplotlib.pyplot as plt

        if output_node is None:
            output_node = self.output_nodes[0]

        if ax is None:
            fig = plt.gcf()
            if len(fig.axes) == 0:
                fig.subplots(1, 1)
            ax = fig.axes[0]

        total = np.sqrt((self.out[output_node] ** 2).sum(1))
        rng = lower * total.min(), upper * total.max()
        noises_to_plot = np.any(self.out[output_node] > rng[0], 0)
        if any(noises_to_plot):
            ax.loglog(self.f, np.abs(self.out[output_node][:, noises_to_plot]))
            ax.loglog(self.f, np.abs(total), c="k", ls="-.", lw=2)
            ax.legend((*np.array(self.noises)[noises_to_plot], "Total"))
            ax.set_ylim(*rng)
        else:
            warn("No noise data to plot in this solution")

        ax.set_ylabel(
            f"ASD [{output_node if not self.scaling else self.scaling}/$\\sqrt{{\\mathrm{{Hz}}}}$]"
        )
        ax.set_xlabel("Frequency [Hz]")


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class NoiseProjection(Action):
    def __init__(self, f, *output_nodes, scaling=None, name="loop"):
        if len(output_nodes) == 0:
            raise ValueError(
                "At least one output node must be specified to compute noise projection to"
            )
        super().__init__(name)
        process = lambda x: x.full_name if type(x) is not str else x

        self.f = f
        self.scaling = process(scaling) if scaling is not None else None
        self.output_nodes = tuple(process(o) for o in output_nodes)

        if len(self.output_nodes) > len(set(self.output_nodes)):
            raise ValueError(
                f"The same output node has been requested multiple times {self.output_nodes}"
            )

    def _do(self, state):
        sol = NoiseProjectionSolution(self.name)
        sol.f = self.f
        sol.output_nodes = self.output_nodes
        sol.scaling = self.scaling
        # create a list of callables func(fsig) to get the ASD noises
        noise_ASDs = {
            name: el.ASD.lambdify(state.model.fsig.f)
            for name, el in state.model.noises.items()
        }
        # labels for noises
        sol.noises = list(el.name for _, el in state.model.noises.items())
        # Keep track of which nodes have what noise injected into them
        noise_node_map = defaultdict(list)
        for el in state.model.noises:
            noise_node_map[el.node.full_name].append(el.name)
        # Collect any extra outputs that should be calculated during the fsig sweep. This
        # is to make efficient use of the filling and solving that this is already doing
        # to extract quantum noise, or others, as needed. Some of these outputs will be
        # signal frequency independant, such as standard shot-noise calculations, so just
        # compute them once.
        # TODO eventually handle qnoised detectors, which are frequency dependant
        fsig_indep_output = []
        for dws in state.sim.readout_workspaces:
            if isinstance(dws, (QShot0Workspace, QShotNWorkspace)):
                added = 0  # don't calculate anything if we aren't modelling the nodes
                # The quantum shot noise detectors will be on the optiical
                # input node, which we can't inject a signal into for computing
                # the noise propagation. Here we need to get the electrical outputs
                # of the readout and just put the noise in there
                for n in dws.owner.signal_nodes:
                    if n.full_name in state.sim.signal.nodes:
                        noise_node_map[n.full_name].append(dws.oinfo.name)
                        added += 1
                if added:
                    fsig_indep_output.append(dws)

        # None actually added
        if len(fsig_indep_output) == 0:
            fsig_indep_output = None

        # Compute all the required transfer functions for noise propagation
        # NOTE: We use _do directly here because we just want to call the action
        # on this state, rather than `run` which will try and create a new state.
        # This is fine, as long as we have requested all the options it needs in
        # _requests. We can't make this frequency response in the init as we do
        # not have the model to grab all the various noise and shot-noise nodes
        self.input_nodes = tuple(noise_node_map.keys())
        sol.freqresp = FrequencyResponse(
            self.f, self.input_nodes, self.output_nodes
        )._do(state, fsig_indep_output)

        # Get any shot noise outputs from the solution
        if fsig_indep_output is not None:
            for dws in fsig_indep_output:
                # Make a simple callable to work wiht the noise ASD functions
                noise_ASDs[dws.oinfo.name] = lambda f: sol.freqresp.extra_outputs[
                    dws.oinfo.name  # noqa: B023
                ]
                sol.noises.append(dws.oinfo.name)

        # get a map from nodes->noise, for noise->node index in output
        inv_noise_node_map = {}
        for k, v in noise_node_map.items():
            for n in v:
                inv_noise_node_map[n] = self.input_nodes.index(k)
        # Convert all the ASDs into PSDs
        sol.PSDs = np.array(
            tuple(np.ones_like(self.f) * fn(self.f) ** 2 for fn in noise_ASDs.values()),
            dtype=float,
        ).T
        # Use this to broadcast from the frequency response output to get the right
        # transfer function for each noise source
        inp_indices = tuple(inv_noise_node_map[name] for name in noise_ASDs.keys())

        # Here we can compute some projection for calculating equivalent noise budgets
        if self.scaling:
            sol.scaling_solution = FrequencyResponse(
                self.f, self.scaling, self.output_nodes, open_loop=True
            )._do(state)
        sol.out = {}
        # compute abs(H)**2 for noise projection of PSDS
        HH = np.zeros(
            (len(self.f), len(self.output_nodes), len(inp_indices)), dtype=float
        )
        np.abs(sol.freqresp.out[:, :, inp_indices], out=HH)
        np.multiply(HH, HH, out=HH)
        # The final index of HH is the output node index, so we can quickly iterate over
        # them here to project the noises
        for i, output_node in enumerate(self.output_nodes):
            # sqrt(output**2/node**2 * node**2/Hz) => output/rtHz
            sol.out[output_node] = np.sqrt(HH[:, i, :] * sol.PSDs)
            if self.scaling:
                # output / scaling
                # sqrt(output**2/node**2 * node**2/Hz) => scaling/rtHz
                sol.out[output_node] /= np.abs(sol.scaling_solution.out[:, 0, :])
        return sol

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append("fsig.f")
        memo["output_nodes"].extend(self.output_nodes)
        if self.scaling:
            memo["input_nodes"].append(self.scaling)
        memo["input_nodes"].extend(
            (el.node.full_name for n, el in model.noises.items())
        )



"""Serial Action."""

import logging

from tqdm.auto import tqdm

import finesse.config
from finesse.solutions import BaseSolution

from .base import Action, convert_str_to_parameter

LOGGER = logging.getLogger(__name__)


class SeriesSolution(BaseSolution):
    """SeriesSolution is a solution object that contains the results from running the
    requested actions by a :class:`.Series`. You can see what solutions are contained by
    printing this object or by inspecting the ``children`` attribute.
    """


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Series(Action):
    """A sequential series of actions to apply during a simulation.

    Parameters
    ----------
    actions : :class:`.Action`
        A collection of Actions to run in series
    flatten : bool
        When True, each action will be stored in the top level of the solution
        tree. When False, each action will be a child of the previous solution
        generated.
    name : str
        Optional name for the solution generated by these actions
    """

    def __init__(self, *actions, flatten=True, name=None):
        super().__init__(name or "series", True)
        self.actions = actions
        self.flatten = flatten

    def _do(self, state):
        rq = self.get_requests(state.model)
        params = tuple(
            convert_str_to_parameter(state.model, _) for _ in rq["changing_parameters"]
        )
        if state.sim is None:
            state.build_model(params, rq["input_nodes"], rq["output_nodes"])

        if self.flatten:
            first = SeriesSolution(self.name)
        else:
            first = None

        curr_sol = None
        pbar = tqdm(
            self.actions,
            disable=len(self.actions) <= 1 or (not finesse.config.show_progress_bars),
            leave=False,
        )

        for action in pbar:
            pbar.set_description_str(self.name)
            next_sol = state.apply(action)
            if self.flatten and next_sol is not None:
                first.add(next_sol)
            else:
                if next_sol and not curr_sol:
                    first = next_sol  # need to return the first one
                if next_sol:
                    if curr_sol:
                        curr_sol.add(next_sol)
                    curr_sol = next_sol
        return first

    def _requests(self, model, memo, first=True):
        for action in self.actions:
            action._requests(model, memo, False)


class ForSolution(BaseSolution):
    """Solution generated by a `For` Action.

    Attributes
    ----------
    parameter : object | str
        Model parameter that was varied by the For action
    values : array_like
        Array of values that were iterated over

    Examples
    --------
    A simple example printing some a varied laser power output:

    >>> import finesse
    >>> model = finesse.Model()
    >>> model.parse('''
    ... l l1
    ... pd P l1.p1.o
    ... ''')
    >>> sol = model.run('for(l1.P, [0, 1, 2, 3], print(l1.P))')
    0.0 W
    1.0 W
    2.0 W
    3.0 W
    >>> print(sol.values)
    [0, 1, 2, 3]
    """


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class For(Action):
    """An action changes a parameter value and runs a set of actions for each value in
    the array. Essentially the same as `Series` combined with `Change`. The parameter
    value is not reset to its original value once this action has finished.

    Generates a :class:`.ForSolution` output when run.

    Examples
    --------
    A simple example printing some a varied laser power output.

    >>> import finesse
    >>> model = finesse.Model()
    >>> model.parse('''
    ... l l1
    ... pd P l1.p1.o
    ... ''')
    >>> sol = model.run('for(l1.P, [0, 1, 2, 3], print(l1.P))')
    0.0 W
    1.0 W
    2.0 W
    3.0 W
    >>> print(sol.values)
    [0, 1, 2, 3]

    Parameters
    ----------
    param : str|Parameter
        Parameter to change in the model
    values : array_like
        Array of values to use for the parameter
    \\*actions : tuple(Action)
        Actions to run for each parameter value.
    """

    def __init__(self, param, values, *actions):
        super().__init__("for", True)
        self.param = param
        self.values = values
        self.actions = actions

    def _do(self, state):
        rq = self.get_requests(state.model)
        for_param = convert_str_to_parameter(state.model, self.param)
        params = tuple(
            [
                for_param,
                *(
                    convert_str_to_parameter(state.model, _)
                    for _ in rq["changing_parameters"]
                ),
            ]
        )
        if state.sim is None:
            state.build_model(params, rq["input_nodes"], rq["output_nodes"])

        first = ForSolution(self.name)
        first.parameter = self.param
        first.values = self.values

        curr_sol = None
        pbar = tqdm(
            self.actions,
            disable=len(self.actions) <= 1 or (not finesse.config.show_progress_bars),
            leave=False,
        )

        for for_param.value in self.values:  # noqa: B007
            state.sim.update_all_parameter_values()
            for action in pbar:
                pbar.set_description_str(self.name)
                next_sol = state.apply(action)
                if next_sol is not None:
                    first.add(next_sol)
                else:
                    if next_sol and not curr_sol:
                        first = next_sol  # need to return the first one
                    if next_sol:
                        if curr_sol:
                            curr_sol.add(next_sol)
                        curr_sol = next_sol
        return first

    def _requests(self, model, memo, first=True):
        memo["changing_parameters"].append(self.param)
        for action in self.actions:
            action._requests(model, memo, False)



import copy
from difflib import unified_diff
import re

from typing import TYPE_CHECKING

from finesse.exceptions import FinesseException
from finesse.env import is_interactive

if TYPE_CHECKING:
    from finesse.components.general import Connector


class MissingReplacementError(FinesseException):
    pass


class ConnectorReplacer:

    def __init__(self, katscript, connector: "Connector") -> None:
        self.katscript: str = katscript
        self.connector: "Connector" = connector

    def replace(
        self,
        sub: str,
        component: str | None = None,
        optical_ports: None | list[str] = None,
        mechanical_ports: None | list[str] = None,
        electrical_ports: None | list[str] = None,
        verbose: bool = False,
    ) -> str:
        # prevents merging two lines together
        if not sub.endswith("\n"):
            sub += "\n"

        # don't override, so you could call this multiple times
        new_kat = copy.copy(self.katscript)

        new_kat = self._replace_definition_line(new_kat, sub)
        new_kat = self._replace_ports_and_nodes(
            new_kat, optical_ports, mechanical_ports, electrical_ports
        )
        new_kat = self._replace_component_name(new_kat, component)

        # TODO maybe not necessary?
        new_kat = new_kat.replace(
            "# Items below could not be matched to original script, or were not present when the model was originally parsed.",
            "",
        )
        new_kat = new_kat.strip()
        if verbose:
            print_diff(self.katscript.strip(), new_kat)
        return new_kat

    def _replace_definition_line(self, old_kat: str, sub: str) -> str:
        # prevent circular import
        from finesse.script.spec import KATSPEC

        aliases = KATSPEC.elements[self.connector.__class__.__name__].aliases
        found = False
        # replace the line constructing the connector we are replacing
        for alias in aliases:
            new_kat = re.sub(
                f"^[ \f\t]*?{alias}[ \f\t]+{self.connector.name}(?=[\\s#]).*?$\n?",
                sub,
                old_kat,
                flags=re.MULTILINE,
            )
            if new_kat != old_kat:
                found = True
                break
        if found is False:
            raise FinesseException(
                f"Could not find definition line for '{self.connector}'"
            )
        return new_kat

    def _replace_ports_and_nodes(
        self, new_kat, optical_ports, mechanical_ports, electrical_ports
    ) -> str:
        # TODO this also replaces ports as part of nodes, which may not be what the
        # user intended?
        connector_ports = (
            self.connector.optical_ports,
            self.connector.mechanical_ports,
            self.connector.electrical_ports,
        )
        replacement_ports = (
            optical_ports,
            mechanical_ports,
            electrical_ports,
        )
        for conn_ports, repl_ports in zip(connector_ports, replacement_ports):
            for i, port in enumerate(conn_ports):
                if port.full_name in new_kat:
                    if repl_ports is None:
                        raise MissingReplacementError(
                            f"No replacement port for '{port.full_name}'"
                        )
                    try:
                        replace_port = repl_ports[i]
                    except IndexError:
                        raise MissingReplacementError(
                            f"No replacement port for '{port.full_name}'"
                        )
                    new_kat = new_kat.replace(port.full_name, replace_port)
        return new_kat

    def _replace_component_name(
        self, new_kat: str, component: str | None = None
    ) -> str:
        # replace references to the component name itself (like in a pendulum)
        # we have to use regex patterns, since, unlike port/node names, a component
        # name can be used inside another name (m m1, m m1_AR)

        lines = new_kat.splitlines(keepends=True)
        patterns = [
            " {} ",  # name surrounded by whitespace
            " {}\n",  # name surrounded by whitespace and newline
            " {}#",  # name directly followed by a comment
            "={} ",  # equal sign and whitespace
            "={}\n",  # equal sign and newline
            "={}#",  # equal sign and comment
            "{}.",  # component parameter being accessed (like in symbolics)
        ]
        new_kat = ""
        for line in lines:
            for pattern in patterns:
                search = pattern.format(self.connector.name)
                if search in line:
                    if component is None:
                        raise ValueError(
                            f"No replacement for '{self.connector.name}' in line '{line}'"
                        )
                    replace = pattern.format(component)
                    line = line.replace(search, replace)
            new_kat += line
        return new_kat


def red(text, html=False):
    if html:
        return f'<span style="color: red">{text}</span>'
    else:
        return f"\033[38;2;255;0;0m{text}\033[38;2;255;255;255m"


def green(text, html=False):
    if html:
        return f'<span style="color: green">{text}</span>'
    else:
        return f"\033[38;2;0;255;0m{text}\033[38;2;255;255;255m"


def print_diff(old: str, new: str) -> None:
    html = is_interactive()
    for line in unified_diff(old.splitlines(), new.splitlines()):
        if line.startswith("-") and not line[1] == "-":
            line = red(line, html)
        elif line.startswith("+") and not line[1] == "+":
            line = green(line, html)
        if html:
            from IPython.display import display, HTML

            display(HTML(line))
        else:
            print(line)



import logging

from libc.stdlib cimport free, calloc

cimport cython

from finesse import BeamParam
from finesse.env import warn
from finesse.frequency cimport Frequency, FrequencyContainer
from finesse.cymath cimport complex_t
from finesse.cymath.math cimport float_eq
from finesse.cymath.complex cimport conj
from finesse.cymath.gaussbeam cimport beam_param, transform_q, inv_transform_q
from finesse.symbols import Constant
from finesse.exceptions import BeamTraceException, NotChangeableDuringSimulation
from finesse.components.modal.cavity cimport CavityWorkspace
from finesse.warnings import CavityUnstableWarning
from finesse.utilities.collections import OrderedSet
from finesse.utilities.cyomp cimport determine_nthreads_even
from finesse.parameter cimport Parameter

from finesse.simulations.workspace cimport ABCDWorkspace
from finesse.element_workspace cimport ElementWorkspace

LOGGER = logging.getLogger(__name__)


cdef class BaseSimulation:
    """This base level class should be inherited by others to perform the exact
    needs of a particular type of simulation. This BaseSimulation class contains
    the methods to store common settings as well as functionality for beam tracing
    through a model, and detecting which beam parameters will be changing.
    Developers should ensure that methods raising `NotImplementedError` are defined
    in any deriving classes.
    """
    def __init__(self, model, unicode name, dict simulation_options):
        self.model = model
        self.name = name
        self.initial_trace_sol = None
        self.trace = NULL
        self.changing_mismatch_couplings = ()
        self.contingent_trace_forests = {}
        self.needs_reflag_changing_q = False
        self.simulation_options = simulation_options
        self.workspace_name_map = {}

    def __dealloc__(self):
        if self.trace:
            free(self.trace)

    def build_carrier_solver(self, nodes, optical_frequencies):
        raise NotImplementedError()

    def build_signal_solver(self, nodes, optical_frequencies, signal_frequencies):
        raise NotImplementedError()

    def build(self):
        cf = self.carrier_frequencies_to_use
        self.carrier = self.build_carrier_solver(self.optical_nodes_to_use, cf)

        if self.compute_signals:
            osf, sf, =self.signal_optical_frequencies_to_use, self.signal_frequencies_to_use
            self.signal = self.build_signal_solver(self.optical_nodes_to_use + self.signal_nodes_to_use, osf, sf)

        self.setup_build()

    def setup_build(self):
        raise NotImplementedError()

    def pre_build(self):
        """
        Pre-build performs some common setup routine that should be applicable to
        any deriving simulation class. This method performs the following tasks:


            * Checks if the signal frequency (fsig) is changing and if its value
              is None. If so, it sets a default value of 1 Hz.
            * Creates a list of all the changing parameters in the simulation.
            * Creates a set of tunable parameters from the changing parameters.
            * Determines if signal computation is required based on the value of
              fsig.f. Sets self.compute_signals
            * Determines if the simulation is modal or not. Sets self.is_modal
            * Initializes the model settings
            * Initializes the simulation configuration data.
            * Generates carrier frequencies based on the model.
            * Initializes the trace forest if the simulation is modal.
            * Determines the changing beam parameters if the simulation is modal.

        """
        if self.model.fsig.f.is_changing and self.model.fsig.f.value is None:
            warn(
                "Signal frequency (fsig) was set to None but simulation needs it. "
                "Setting default value of 1 Hz"
            )
            self.model.fsig.f.value = 1

        self.changing_parameters = OrderedSet(
            (p for p in self.model.all_parameters if p.is_changing)
        )

        for p in self.changing_parameters:
            if not p.changeable_during_simulation:
                raise NotChangeableDuringSimulation(
                    f"Parameter {p} cannot be changed during a simulation. Loop over this parameter changed with a Python for loop instead."
                )

        self.tunable_parameters = OrderedSet(
            p for p in self.changing_parameters if p.is_tunable
        )
        self.compute_signals = self.model.fsig.f.value is not None
        self.is_modal = self.model.is_modal
        self.initialise_model_settings()
        self.initialise_sim_config_data()

        self.optical_nodes_to_use = self.model.optical_nodes
        self.carrier_frequencies_to_use = self.generate_carrier_frequencies()

        if self.compute_signals:
            self.signal_nodes_to_use = list(self.model.get_active_signal_nodes())
            self.signal_optical_frequencies_to_use, self.signal_frequencies_to_use = self.generate_signal_frequencies(
                self.optical_nodes_to_use + self.signal_nodes_to_use,
                self.carrier_frequencies_to_use
            )

        self.initialise_trace_forest(self.model.optical_nodes)

        for el in self.model.elements.values():
            el._setup_changing_params()


    def post_build(self):
        """Post build calls functions that should be called after everything else has been
        built. For example, detectors that might rely on components being set up.
        """
        if self.carrier is None:
            raise RuntimeError("No carrier simulation was made")
        if self.compute_signals and self.signal is None:
            raise RuntimeError("No carrier simulation was made")

        self.setup_output_workspaces()

    def unbuild(self):
        if self.carrier is not None:
            self.carrier.destruct()
        if self.signal is not None:
            self.signal.destruct()

        for p in self.changing_parameters:
            # Reset all changing parameters so their type can change again
            (<Parameter>p).__disable_state_type_change = False

        # Code below can be used in debug mode to determine if anyone is keeping any
        # references to this matrix object, meaning its memory can't be freed.
        # This takes ~20ms to do so makes a difference for quick models. Maybe we need
        # a debug mode

        #_ref = self._M
        #self._M = None

        # refs = gc.get_referrers(_ref)
        # Nref = len(refs)
        # if Nref > 0:
        #     warn(
        #         f"Something other than the Simulation object (N={Nref}) is keeping"
        #         f" a reference to the matrix object {repr(self._M)}."
        #         " Could lead to excessive memory usage if not released."
        #     )
        #     for _ in refs:
        #         warn(f" - {repr(_)}")
        #del _ref

    cpdef update_cavities(self):
        cdef CavityWorkspace ws
        for ws in self.cavity_workspaces.values():
            ws.update()

    cpdef update_map_data(self):
        """This will cycle through each map being used and if
        it is defined by a function it will be evaluated again.
        """
        cdef ABCDWorkspace ws
        # This could be made more efficient by just storing the
        # a list of those with changing maps
        for ws in self.to_scatter_matrix_compute:
            ws.update_map_data()

    cdef int set_gouy_phases(self) except -1:
        cdef ABCDWorkspace ws
        cdef int rtn

        for ws in self.gouy_phase_workspaces:
            if ws.fn_gouy_c is not None:
                rtn = ws.fn_gouy_c.func(ws)
            elif ws.fn_gouy_py is not None:
                rtn = ws.fn_gouy_py(ws)
            if rtn:
                return rtn

        return 0

    cpdef modal_update(self):
        """Updates HOM related dependencies / properties of the model.

        These updates are as follows:

         * Execute a beam trace on the changing trace trees
         * Computes the changing scattering matrices
         * Calculates the Gouy phase of Spaces and Laser power coefficients

        Returns
        -------
        validity : bool
            True if the modal update was successful, or False if an unstable
            cavity combination prevented a beam trace from being performed.
        """
        # Evaluate changing properties of cavity workspaces
        self.update_cavities()

        if self.retrace:
            if not self.trace_beam():
                return False

        # Update the changing Gouy phases at spaces
        # and TEM Gouy phases at lasers
        self.set_gouy_phases()

        return True


    cdef int _determine_changing_beam_params(
        self, TraceForest forest=None, bint set_tree_node_ids=True,
    ):
        if self.trace == NULL:
            raise RuntimeError("trace is NULL")

        cdef:
            Py_ssize_t i
            Py_ssize_t num_nodes = len(self.model.optical_nodes)
            ABCDWorkspace kws

        # Re-set all beam parameter changing flags to false initially
        for i in range(num_nodes):
            self.trace[i].is_fixed = True

        if self.retrace:
            LOGGER.info("Flagging changing beam parameters.")
            # Prepare the forest for simulation by setting all the node_id attributes
            # and flag the corresponding self.trace entries as changing
            self._setup_trace_forest(forest, set_tree_node_ids)

        # Now tell each knm workspace whether it is changing or not
        # so that only changing scattering matrices get recomputed
        # from here on
        if self.to_scatter_matrix_compute is not None:
            for kws in self.to_scatter_matrix_compute:
                kws.flag_changing_beam_parameters(self.changing_mismatch_edges)

        return 0

    def is_component_in_mismatch_couplings(self, comp):
        """Determines whether the connector `comp` is associated with any
        of the node couplings in the stored changing mismatch couplings.

        .. note::

            This method can be replaced if connectors eventually use more
            granular refill flags --- i.e. per coupling refill flags. Then
            the check for refilling that coupling can simply include the
            condition ``(from_node, to_node) in sim.changing_mismatch_couplings``.
        """
        return any(node.component is comp for node, _ in self.changing_mismatch_couplings)

    cdef _setup_trace_forest(self, TraceForest forest=None, bint set_tree_node_ids=True):
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

        if forest is None:
            forest = self.trace_forest

        for tree_idx in range(forest.size()):
            tree = forest.forest[tree_idx]
            self._setup_single_trace_tree(tree, set_tree_node_ids)

    cdef _setup_single_trace_tree(self, TraceTree tree, bint set_tree_node_ids=True):
        cdef:
            TraceTree ltree = tree.left
            TraceTree rtree = tree.right

        # Only ever need to do this once, so avoid repeating when reflagging
        # changing beam params after exiting unstable cavity regions
        if set_tree_node_ids:
            tree.node_id = self.trace_node_index[tree.node]
            tree.opp_node_id = self.trace_node_index[tree.node.opposite]

        self.trace[tree.node_id].is_fixed = False
        self.trace[tree.opp_node_id].is_fixed = False

        if ltree is not None:
            self._setup_single_trace_tree(ltree)
        if rtree is not None:
            self._setup_single_trace_tree(rtree)

    cdef tuple _find_new_unstable_cavities(self) :
        cdef:
            Py_ssize_t tree_idx
            TraceTree tree

            CavityWorkspace cav_ws
            bint source_is_cav
            double gx, gy

            list ch_unstable_cavities = []

        for tree_idx in range(self.trace_forest.size()):
            tree = self.trace_forest.forest[tree_idx]

            if tree.is_source:
                cav_ws = self.cavity_workspaces.get(tree.dependency)
                source_is_cav = cav_ws is not None
                if source_is_cav: # Tree is an internal cavity tree
                    # The geometrically changing cavity has become unstable
                    # so inform that this is the case
                    if not cav_ws.is_stable:
                        ch_unstable_cavities.append(cav_ws.owner)

                        gx = cav_ws.gx
                        gy = cav_ws.gy
                        if float_eq(gx, gy):
                            warn(
                                f"Cavity {repr(tree.dependency.name)} is unstable with "
                                f"g = {gx}",
                                CavityUnstableWarning
                            )
                        else:
                            warn(
                                f"Cavity {repr(tree.dependency.name)} is unstable with "
                                f"gx = {gx}, gy = {gy}",
                                CavityUnstableWarning
                            )

        if not ch_unstable_cavities:
            return ()

        # Return tuple of the unstable cavities sorted by name so that
        # all permutations of the same combination of cavities give same
        # tuple --- important for look-ups in contingent_trace_forests dict
        return tuple(sorted(ch_unstable_cavities, key=lambda x: x.name))

    cdef TraceForest _initialise_contingent_forest(self, tuple unstable_cavities) :
        cdef TraceForest contingent_forest = TraceForest(self.model, self.trace_forest.symmetric)
        cdef TraceForest model_trace_forest = self.model.trace_forest
        cdef list order = model_trace_forest.dependencies.copy()
        for uc in unstable_cavities:
            order.remove(uc)

        # If there are no dependencies left after disabling the
        # unstable cavities then a beam trace cannot be performed
        # at this data point so no forest can be planted
        if not order:
            warn(
                "Cannot build a contingent trace forest as the simulation "
                "is in a regime with no stable cavities nor Gauss objects.",
                CavityUnstableWarning
            )
            return None

        contingent_forest.plant(order)

        if self._determine_changing_beam_params(contingent_forest):
            return None

        return contingent_forest

    @cython.initializedcheck(False)
    cdef void _propagate_trace(self, TraceTree tree, bint symmetric) noexcept:
        cdef:
            TraceTree ltree = tree.left
            TraceTree rtree = tree.right

            const NodeBeamParam* q1 = &self.trace[tree.node_id]
            complex_t qx1 = q1.qx.q
            complex_t qy1 = q1.qy.q
            complex_t qx2, qy2

        if ltree is not None:
            # For non-symmetric traces we have some special checks
            # to do on trees which couldn't be reached from the
            # other dependency trees. Note these are only performed
            # on the left tree; see TraceForest._add_backwards_nonsymm_trees
            # for details.
            if symmetric or (not tree.do_nonsymm_reverse and not tree.do_inv_transform):
                qx2 = transform_q(tree.left_abcd_x, qx1, tree.nr, ltree.nr)
                qy2 = transform_q(tree.left_abcd_y, qy1, tree.nr, ltree.nr)
            elif tree.do_inv_transform:
                # Can't reach tree directly but there is a coupling from ltree.node
                # to tree.node so apply the inverse abcd law to get correct q
                qx2 = inv_transform_q(tree.left_abcd_x, qx1, tree.nr, ltree.nr)
                qy2 = inv_transform_q(tree.left_abcd_y, qy1, tree.nr, ltree.nr)
            else:
                # Really is no way to get to the node (no coupling from ltree.node to
                # tree.node) so only option now is to reverse q for ltree node entry
                qx2 = -conj(qx1)
                qy2 = -conj(qy1)

            self.trace[ltree.node_id].qx.q = qx2
            self.trace[ltree.node_id].qy.q = qy2
            if symmetric:
                self.trace[ltree.opp_node_id].qx.q = -conj(qx2)
                self.trace[ltree.opp_node_id].qy.q = -conj(qy2)

            self._propagate_trace(ltree, symmetric)

        if rtree is not None:
            qx2 = transform_q(tree.right_abcd_x, qx1, tree.nr, rtree.nr)
            qy2 = transform_q(tree.right_abcd_y, qy1, tree.nr, rtree.nr)

            self.trace[rtree.node_id].qx.q = qx2
            self.trace[rtree.node_id].qy.q = qy2
            if symmetric:
                self.trace[rtree.opp_node_id].qx.q = -conj(qx2)
                self.trace[rtree.opp_node_id].qy.q = -conj(qy2)

            self._propagate_trace(rtree, symmetric)

    cpdef trace_beam(self) :
        """Traces the beam through the paths which are dependent upon changing
        geometric parameter(s).

        This method will modify those entries in the ``self.trace`` C array
        which were previously determined to have changing beam parameter values.

        Returns
        -------
        validity : bool
            True if the tracing was successful, or False if an unstable
            cavity combination prevented a beam trace from being performed.

        Raises
        ------
        ex : :class:`.BeamTraceException`
            If the ``"unstable_handling"`` entry of the associated
            :attr:`.Model.sim_trace_config` dict is ``"abort"`` and
            any unstable cavities were encountered.
        """
        cdef:
            TraceTree tree
            Py_ssize_t tree_idx

            CavityWorkspace cav_ws
            bint source_is_cav

            complex_t qx_src, qy_src

            # The actual trace forest which gets traced. In most circumstances
            # this will be self.trace_forest but if changing cavities enter an
            # unstable regime then this will be temporarily swapped out for the
            # contingent forest for this data point (see below).
            TraceForest trace_forest

            # Objects necessary for dealing with newly unstable cavities
            tuple ch_unstable_cavities
            TraceForest contingent_forest

        # No changing beam parameters, do nothing
        if self.trace_forest.empty():
            return True

        # First we loop over the source trees and find any changing
        # cavities which have become unstable
        ch_unstable_cavities = self._find_new_unstable_cavities()

        # If we did find any newly unstable cavities then the current trace_forest
        # is invalidated at the current data point so we must build a new forest with
        # the unstable cavities disabled
        # NOTE (sjr) Don't worry about increased Python interaction in
        #            this block as this will rarely be executed anyway
        if ch_unstable_cavities:
            unstable_handling = self.model.sim_trace_config["unstable_handling"]

            # Abort simulation if tracing config set-up to do so when
            # encountering any unstable cavities
            if unstable_handling == "abort":
                raise BeamTraceException(
                    "Aborting simulation due to presence of unstable cavities: "
                    f"{','.join([cav.name for cav in ch_unstable_cavities])}"
                )
            # Or if tracing config is set-up to abort only the retrace when
            # any unstable cavities encountered, flag this to notify that
            # appropriate detector outputs should be masked
            if unstable_handling == "mask":
                LOGGER.info(
                    "Aborting retrace as simulation tracing configuration is set-up "
                    "to mask at any data point(s) where unstable cavities occur."
                )
                return False

            LOGGER.info(
                "Attempting to use a contingent trace forest due "
                "to the presence of unstable cavities"
            )
            # Look-up the combination of unstable cavities to see if a
            # contingent forest was already built from this
            contingent_forest = self.contingent_trace_forests.get(ch_unstable_cavities)

            # No previous forest built from the given combination of disabled
            # unstable cavities so need to build one here
            if contingent_forest is None:
                LOGGER.debug(
                    "For unstable cavity combination %s no cached contingent "
                    "trace forest found, now attempting to build a new one...",
                    [uc.name for uc in ch_unstable_cavities],
                )
                contingent_forest = self._initialise_contingent_forest(ch_unstable_cavities)
                # If there are no dependencies left after disabling the
                # unstable cavities then a beam trace cannot be performed
                # at this data point so inform of this on return
                if contingent_forest is None:
                    return False

                # Cache the contingent forest for this combination of unstable
                # cavities as these typically occur in blocks (or across strides)
                # of data points so we don't want to keep rebuilding the same
                # contingency forests for identical unstable cavity combos
                self.contingent_trace_forests[ch_unstable_cavities] = contingent_forest
            else:
                LOGGER.debug(
                    "For unstable cavity combination %s found and using "
                    "cached contingent trace forest:%s",
                    [uc.name for uc in ch_unstable_cavities],
                    contingent_forest
                )

            # Make sure only the correctly changing beam parameters, according
            # to self.trace_forest, get reflagged when exiting from the unstable
            # region again
            self.needs_reflag_changing_q = True

            # Use the contingent forest for this data point
            trace_forest = contingent_forest

        # Otherwise we just use the standard changing trace forest of the simulation
        else:
            trace_forest = self.trace_forest

            # If we've just exited an unstable cavity region where a contingent trace
            # forest was being used, then we need to reflag the beam parameters which
            # are changing
            if self.needs_reflag_changing_q:
                if self._determine_changing_beam_params(forest=None, set_tree_node_ids=False):
                    return False
                self.needs_reflag_changing_q = False

        # Now do the actual beam tracing by simply traversing the forest
        # and propagating the beam through each tree
        for tree_idx in range(trace_forest.size()):
            tree = trace_forest.forest[tree_idx]

            if tree.is_source:
                cav_ws = self.cavity_workspaces.get(tree.dependency)
                source_is_cav = cav_ws is not None

                if not source_is_cav: # Source tree is from a Gauss
                    # TODO (sjr) Should probably make some workspace for Gauss objects
                    #            which then uses cy_expr's for evaluating these things
                    #            if they're symbolic. But for now this will do.
                    qx_src = complex(tree.dependency.qx.q)
                    qy_src = complex(tree.dependency.qy.q)
                else: # Source tree is from a Cavity
                    qx_src = cav_ws.qx
                    qy_src = cav_ws.qy

                self.trace[tree.node_id].qx.q = qx_src
                self.trace[tree.node_id].qy.q = qy_src
                if trace_forest.symmetric:
                    self.trace[tree.opp_node_id].qx.q = -conj(qx_src)
                    self.trace[tree.opp_node_id].qy.q = -conj(qy_src)

            self._propagate_trace(tree, trace_forest.symmetric)

        return True

    cpdef run_carrier(self) :
        """Runs the carrier matrix solver for the current state of the model.
        This will update all the C based structs with the current model state so
        that filling and calculations can be performed.

        Returns
        -------
        validity : bool
            True if this was a valid run, or False if a recoverable error occurred
            which results in the output being invalid for this call.
        """
        # NOTE (sjr) Just updating all parameter values on each call to run for
        #            now. This may not be the most optimal thing to do, but it
        #            avoids duplicating these parameter update calls in different
        #            places (e.g. refill, compute_knm_matrices, set_gouy_phases) and
        #            should be safe in that no parameters get accidentally missed at
        #            any data point.
        # ddb - this just updates everything, even things that are not changing as
        # it acts on all the workspaces, probably not the best idea
        self.update_all_parameter_values()

        # Update HOM stuff
        if self.is_modal:
            # Immediately return if invalid beam trace region encountered
            # no need to go ahead and fill or solve as they won't be used
            if not self.modal_update():
                return False

        self.carrier.run()

        return True

    cpdef run_signal(self, solve_noises=True) :
        """Runs the signal matrix solver for the current state. This function should assume that
        a call to the `run_carrier` method has preceeded it. Many modal and parameter updates
        should happen in there already, so do not need to be repeated here.
        """
        self.model_settings.fsig = float(self.model.fsig.f.value)
        # Probably some other preparatory stuff needs to go here in the future
        self.signal.run()

        # Then ask components for their noise contributions
        if solve_noises and self.signal.noise_sources:
            self.signal.fill_noise_inputs()
            self.signal.solve_noises()

    def setup_output_workspaces(self):
        from finesse.detectors.general import NoiseDetector
        from finesse.components.readout import _Readout
        from finesse.detectors.workspace import DetectorWorkspace

        # Once the simulations are started we can tell all the detectors to
        # prepare themselves
        self.detector_workspaces = []
        self.readout_workspaces = []

        for rd in self.model.get_elements_of_type(_Readout):
            # Readouts can emulate multiple detectors, so here we
            # get a collection of them depending on what the readout
            # is doing and add them to the list
            for output, ws in rd._get_output_workspaces(self).items():
                if not isinstance(ws, DetectorWorkspace):
                    raise TypeError(f"Readout detector ({rd}) workspace ({output.ws}) not a DetectorWorkspace type")
                self.readout_workspaces.append(ws)
                self.workspace_name_map[output.name] = ws
                if rd.output_detectors:
                    self.detector_workspaces.append(ws)

        for det in self.model.detectors:
            ws = det._get_workspace(self)
            self.workspace_name_map[det.name] = ws

            if ws is not None:
                if not isinstance(ws, DetectorWorkspace):
                    raise TypeError(f"Detector ({det}) workspace ({ws}) not a DetectorWorkspace type")

                self.detector_workspaces.append(ws)
                if self.signal and isinstance(ws.owner, NoiseDetector):
                    self.signal.workspaces.noise_detectors.append(ws)

        for _ in self.detector_workspaces:
            _.compile_cy_exprs()

        if self.signal:
            self.signal.workspaces.detector_list_to_C()

    def __enter__(self):
        self.pre_build()
        self.build()
        self.post_build()

    def __exit__(self, type_, value, traceback):
        self.unbuild()

    cpdef update_all_parameter_values(self) :
        """Loops through all workspaces to update the C structs so they
        represent the current model element parameter values.
        """
        cdef:
            ElementWorkspace ws

            Py_ssize_t i
            # TODO (sjr) Should probably cache these or move away from
            #            lists for best performance
            Py_ssize_t Ncws = len(self.workspaces)

        for i in range(Ncws):
            ws = self.workspaces[i]
            ws.update_parameter_values()

    def get_q(self, node):
        """Returns a tuple of (qx, qy) for a given node. The returned
        value is only valid until this simulations trace forest has
        been updated.

        Parameters
        ----------
        node : :class:`finesse.components.OpticalNode`
            Node to get beam parameters at

        Returns
        -------
        (qx, qy)
            Tuple of x and y beam parameters
        """
        cdef NodeBeamParam *nodebp

        idx = self.trace_node_index[node]
        if idx >= 0 and idx < len(self.model.optical_nodes):
            nodebp = &self.trace[idx]
            return (
                BeamParam(q=nodebp.qx.q, nr=nodebp.qx.nr, wavelength=nodebp.qx.wavelength),
                BeamParam(q=nodebp.qy.q, nr=nodebp.qy.nr, wavelength=nodebp.qy.wavelength),
            )
        else:
            raise IndexError("Node index is not in simulation")

    cdef initialise_trace_forest(self, optical_nodes) :
        cdef TraceForest model_trace_forest
        cdef double nr

        self.trace_node_index = {n: i for i, n in enumerate(optical_nodes)}

        # Before we setup the workspaces some initial beam trace must be done
        # so that workspaces can initialise themselves
        if self.is_modal:
            # Make sure the model trace forest gets re-planted
            # when building a new simulation
            self.model._rebuild_trace_forest = True
            LOGGER.info(
                "Performing initial beam trace with configuration options:\n    %s",
                self.model.sim_trace_config,
            )
            # Plant the model trace_forest and execute initial beam trace
            self.initial_trace_sol = self.model.beam_trace(**self.model.sim_initial_trace_args)
            model_trace_forest = self.model.trace_forest
            self.nodes_with_changing_q = model_trace_forest.get_nodes_with_changing_q()
            self.changing_mismatch_edges = OrderedSet()

            LOGGER.info(
                "Nodes with changing q during simulation:\n    %s",
                self.nodes_with_changing_q,
            )
            self.retrace = self.model.sim_trace_config["retrace"]
            self.trace = <NodeBeamParam*> calloc(len(optical_nodes), sizeof(NodeBeamParam))
            if not self.trace:
                raise MemoryError()

            for i, n in enumerate(optical_nodes):
                qx, qy = self.initial_trace_sol[n]
                nr = qx.nr
                self.trace[i] = NodeBeamParam(
                    beam_param(qx.q, nr, self.model_settings.lambda0),
                    beam_param(qy.q, nr, self.model_settings.lambda0),
                    n in self.nodes_with_changing_q
                )

            if self.retrace:
                # Construct the forest of changing trace trees - it's important
                # that this is done before initialising connector workspaces as
                # they need the changing forest to be present for refill flag
                self.trace_forest = model_trace_forest.make_changing_forest()
                self.retrace &= not self.trace_forest.empty()

                if self.retrace:
                    LOGGER.info(
                        "Determined changing trace trees:%s", self.trace_forest
                    )
                    # Get the nodes at which trees of the changing forest intersect
                    # with trees of the full forest which have different trace
                    # dependencies. These couplings will have potentially changing
                    # mode mismatches during the simulation.
                    self.changing_mismatch_couplings = self.trace_forest.find_potential_mismatch_couplings(
                        model_trace_forest
                    )
                    # The above returns node objects. It is more useful to also store
                    # the equivalent simulation specific "node IDs" which is just an
                    # integer and used more in the cythonised code as some nodes get
                    # dropped when not used in a simulation.
                    # Optical nodes for signal and carrier all have the same IDs
                    for i, (ni, no) in enumerate(self.changing_mismatch_couplings):
                        self.changing_mismatch_edges.add(
                            tuple((self.trace_node_index[ni], self.trace_node_index[no]))
                        )

                    if self.changing_mismatch_couplings:
                        LOGGER.info(
                            "Found changing mismatched node couplings: %s",
                            [f"{n1.full_name} -> {n2.full_name}"
                            for n1, n2 in self.changing_mismatch_couplings]
                        )
            else:
                self.trace_forest = TraceForest(self.model, self.model.sim_trace_config["symmetric"])
        else:
            # just make an empty TraceForest
            self.trace_forest = TraceForest(self.model, self.model.sim_trace_config["symmetric"])

    cdef initialise_model_settings(self) :
        self.model_settings = self.model._settings

        if self.model.fsig.f.value is None:
            self.model_settings.fsig = 0
        else:
            self.model_settings.fsig = float(self.model.fsig.f.value)

    cdef initialise_sim_config_data(self) :
        # Nominal number of threads will be Nhoms / 10
        self.config_data.nthreads_homs = determine_nthreads_even(self.model_settings.num_HOMs, 10)

        LOGGER.info("Using %d threads for HOM parallel loops.", self.config_data.nthreads_homs)

    def generate_carrier_frequencies(self):
        """Returns a list of Frequency objects that the model has requested"""
        from finesse.frequency import Frequency, generate_frequency_list
        if len(self.model._frequency_generators) == 0:
            # Nothing in the model is generating a carrier frequency. Typical
            # situation is when no laser is included for signal modelling.
            # Simple solution, just use a default 0Hz
            frequencies_to_use = [Constant(0.0)]
        else:
            frequencies_to_use = generate_frequency_list(self.model)

        carrier_frequencies = list()

        LOGGER.info("Generating simulation with user carrier frequencies %s", self.model.frequencies)

        for i, f in enumerate(self.model.frequencies):
            try:
                f_name = str(f.eval(keep_changing_symbols=True))
            except AttributeError:
                f = Constant(f)
                f_name = str(f)

            carrier_frequencies.append(Frequency(f_name, f, index=i))

        N = len(self.model.frequencies)
        LOGGER.info("Generating simulation with carrier frequencies %s", frequencies_to_use)
        for i, f in enumerate(frequencies_to_use):
            carrier_frequencies.append(
                Frequency(
                    str(f.eval(keep_changing_symbols=True)),
                    f,
                    index=N+i
                )
            )

        fcnt = FrequencyContainer(carrier_frequencies)
        return fcnt

    def generate_signal_frequencies(self, nodes, FrequencyContainer carrier_optical_frequencies):
        """Generates the optical, mechanical, and electrical frequencies that should be
        modelled by the signal simulation.
        """
        from finesse.components.node import NodeType

        optical_frequencies = [] # All optical frequencies are present at all nodes
        # elec and mech can have different frequencies on a per node basis
        signal_frequencies = {}

        for i, f in enumerate(carrier_optical_frequencies.frequencies):
            fp = f.f + self.model.fsig.f.ref
            fm = f.f - self.model.fsig.f.ref

            optical_frequencies.append(
                Frequency(str(fp.eval(keep_changing_symbols=True)),
                            fp, index=2*i, audio_order=1,
                            audio=True, audio_carrier_index=i,
                            audio_carrier_object=f)
            )
            optical_frequencies.append(
                Frequency(str(fm.eval(keep_changing_symbols=True)),
                            fm, index=2*i+1, audio_order=-1,
                            audio=True, audio_carrier_index=i,
                            audio_carrier_object=f)
            )

        fcnt = FrequencyContainer(optical_frequencies, carrier_cnt=carrier_optical_frequencies)

        # Audio matrix frequencies are more complicated as they can have multiple frequencies
        # in mechanical and electrical, on a per-node basis...
        fsig = FrequencyContainer(
            (Frequency("fsig", self.model.fsig.f.ref, index=0), )
        )

        for node in nodes:
            if node.type == NodeType.OPTICAL:
                continue

            #-----------------------------------------------------------------------------------
            # Mechanical frequencies
            #-----------------------------------------------------------------------------------
            # By default mechanical frequencies just have a single frequency at the Model.fsig.f
            # However for more complicated systems we can have multiple frequencies.
            elif node.type == NodeType.MECHANICAL:
                fs = []
                freqs = node.frequencies
                if len(freqs) == 1 and freqs[0] == self.model.fsig.f.ref:
                    # Most components will just use a single fsig so reuse same object
                    # for efficient filling later
                    signal_frequencies[node] = fsig
                else:
                    for i, sym in enumerate(node.frequencies):
                        fs.append(
                            Frequency(
                                str(fm.eval(keep_changing_symbols=True)),
                                sym,
                                index=i,
                            )
                        )
                    signal_frequencies[node] = FrequencyContainer(tuple(fs))

            #-----------------------------------------------------------------------------------
            # Electrical frequencies
            #-----------------------------------------------------------------------------------
            elif node.type == NodeType.ELECTRICAL:
                signal_frequencies[node] = fsig
            else:
                raise ValueError("Unexpected")

        return fcnt, signal_frequencies



import pytest

from finesse import Model
from finesse.exceptions import FinesseException
from finesse.components import ReadoutDC, ReadoutRF, Mirror
from finesse.components.readout import _Readout
from finesse.components.node import Port, Node


# https://gitlab.com/ifosim/finesse/finesse3/-/issues/654
def test_double_connection_raises():
    m = Model()
    m1 = m.add(Mirror("m1"))
    readout = m.add(ReadoutDC(name="readout", optical_node=m1.p2.o))
    with pytest.raises(FinesseException):
        m.connect(m1.p2, readout.p1)


def test_implicit_connection_raises():
    m = Model()
    m1 = m.add(Mirror("m1"))
    readout = m.add(ReadoutDC(name="readout", optical_node=m1.p2.o))
    with pytest.raises(FinesseException):
        m.connect(m1.p1, readout.p1)


@pytest.fixture(
    params=(ReadoutDC, ReadoutRF),
)
def readout_cls(request) -> _Readout:
    return request.param


def test_optical_node(readout_cls):
    model = Model()
    m1 = Mirror("m1")
    model.add(m1)
    readout = readout_cls("readout", optical_node=m1.p2.o)
    assert readout.p1.i is m1.p2.o
    assert readout.p1.o is m1.p2.i


def test_optical_node_with_port(readout_cls):
    model = Model()
    m1 = Mirror("m1")
    model.add(m1)
    with pytest.warns():
        readout = readout_cls("readout", optical_node=m1.p2)
    # when passing in a port, should grab the first node, which is input
    assert readout.p1.i is m1.p2.i
    assert readout.p1.o is m1.p2.o


def test_optical_node_none(readout_cls):
    readout = readout_cls("readout")
    assert isinstance(readout.p1, Port)
    assert isinstance(readout.p1.i, Node)
    assert isinstance(readout.p1.o, Node)


def test_optical_wrong_type(readout_cls):
    model = Model()
    m1 = Mirror("m1")
    model.add(m1)
    with pytest.raises(TypeError):
        readout_cls("readout", optical_node=m1)



import logging
import warnings

from .style import use
from ..env import warn

LOGGER = logging.getLogger(__name__)

DEFAULT_FORMATS = ("svg",)

MOCK_IN_IPYTHON = False


def _in_ipython():
    """Checks whether the current script is running under IPython.

    Returns
    -------
    bool
        True if __IPYTHON__ is defined, otherwise False.
    """
    try:
        __IPYTHON__
    except NameError:
        if MOCK_IN_IPYTHON:
            warnings.warn(
                "Mocking ipython environment, should only be done in testsuite",
                stacklevel=2,
            )
            return True
        return False
    else:
        return True


def init(mode="display", dpi=None, fmts=None):
    """Sets up default plotting parameters for a desired display mode.

    Parameters
    ----------
    mode : str
        Display mode to use, either 'display' or 'paper'.

    dpi : int, optional
        DPI (Dots per inch) to display and save figures with. Defaults to
        current setting.

    fmts : list of str, optional
        List of image formats to allow for display when running under IPython; defaults to
        ``DEFAULT_FORMATS``.
    """
    import matplotlib as mpl

    if fmts is None:
        fmts = DEFAULT_FORMATS

    if _in_ipython():
        try:
            from matplotlib_inline.backend_inline import set_matplotlib_formats

            ipy = get_ipython()
            try:
                ipy.run_line_magic("matplotlib", "inline")
            except KeyError:
                try:
                    ipy.run_line_magic("matplotlib", "qt")
                except KeyError:
                    warn("Could not set matplotlib backend. Tried inline and Qt.")
            finally:
                set_matplotlib_formats(*fmts)
        except NameError:
            pass

    # TODO (sjr) handle plotting options in user config

    if mode == "display":
        use(["default"])
    elif mode == "paper":
        if not _in_ipython():
            mpl.use("pgf")
        use(["default", "paper"])
    else:
        raise (BaseException("Plotting mode must be either 'display' or 'paper'."))

    if dpi is not None:
        dpi = int(dpi)
        mpl.rcParams.update({"figure.dpi": dpi})
        mpl.rcParams.update({"savefig.dpi": dpi})
    # always apply white background and remove transparency. Make plots readable
    # when using dark background/themes in jupyter notebooks and such
    mpl.rcParams.update(
        {
            "figure.facecolor": (1.0, 1.0, 1.0, 1.0),
        }  # red   with alpha = 30%
    )


def add_colorbar(im, **kwargs):
    """Adds a vertical color bar to an image plot.

    Parameters
    ----------
    im : :class:`matplotlib.image.AxesImage`

      An image plot, e.g. return value of a call to ``plt.imshow()``.

    Returns
    -------
    Handle to the colorbar.
    """
    ## Credit to: https://joseph-long.com/writing/colorbars/
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    import matplotlib.pyplot as plt

    last_axes = plt.gca()
    ax = im.axes
    fig = ax.figure
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    cbar = fig.colorbar(im, cax=cax, **kwargs)
    plt.sca(last_axes)
    return cbar



import pytest
from finesse.script import KATSPEC, unparse


@pytest.mark.parametrize(
    "kwargs,directive_defaults,argument_defaults,prefer_keywords,expected",
    (
        ## modes only
        # empty
        ({}, False, False, False, ""),
        ({}, False, False, True, ""),
        ({}, False, True, False, ""),
        ({}, False, True, True, ""),
        ({}, True, False, False, "modes()"),
        ({}, True, False, True, "modes()"),
        ({}, True, True, False, "modes(off)"),
        ({}, True, True, True, "modes(modes=off)"),
        # string
        ({"modes": "off"}, False, False, False, ""),
        ({"modes": "off"}, False, False, True, ""),
        ({"modes": "off"}, False, True, False, ""),
        ({"modes": "off"}, False, True, True, ""),
        ({"modes": "off"}, True, False, False, "modes()"),
        ({"modes": "off"}, True, False, True, "modes()"),
        ({"modes": "off"}, True, True, False, "modes(off, none, none, none)"),
        (
            {"modes": "off"},
            True,
            True,
            True,
            "modes(modes=off, maxtem=none, include=none, remove=none)",
        ),
        # array
        ({"modes": [[0, 1], [1, 0]]}, False, False, False, "modes([[0, 1], [1, 0]])"),
        (
            {"modes": [[0, 1], [1, 0]]},
            False,
            False,
            True,
            "modes(modes=[[0, 1], [1, 0]])",
        ),
        (
            {"modes": [[0, 1], [1, 0]]},
            False,
            True,
            False,
            "modes([[0, 1], [1, 0]], none, none, none)",
        ),
        (
            {"modes": [[0, 1], [1, 0]]},
            False,
            True,
            True,
            "modes(modes=[[0, 1], [1, 0]], maxtem=none, include=none, remove=none)",
        ),
        ## maxtem only
        ({"maxtem": 2}, False, False, False, "modes(none, 2)"),
        ({"maxtem": 2}, False, False, True, "modes(maxtem=2)"),
        ({"maxtem": 2}, False, True, False, "modes(none, 2, none, none)"),
        (
            {"maxtem": 2},
            False,
            True,
            True,
            "modes(modes=none, maxtem=2, include=none, remove=none)",
        ),
        ## modes and maxtem
        # string mode, numeric maxtem
        ({"modes": "even", "maxtem": 4}, False, False, False, "modes(even, 4)"),
        (
            {"modes": "even", "maxtem": 4},
            False,
            False,
            True,
            "modes(modes=even, maxtem=4)",
        ),
        (
            {"modes": "even", "maxtem": 4},
            False,
            True,
            False,
            "modes(even, 4, none, none)",
        ),
        (
            {"modes": "even", "maxtem": 4},
            False,
            True,
            True,
            "modes(modes=even, maxtem=4, include=none, remove=none)",
        ),
    ),
)
def test_modes(
    model,
    kwargs,
    directive_defaults,
    argument_defaults,
    prefer_keywords,
    expected,
):
    adapter = KATSPEC.commands["modes"]

    if kwargs:
        adapter.setter(model, ((), kwargs))

    dump = next(iter(adapter.getter(adapter, model)))
    script = unparse(
        dump,
        directive_defaults=directive_defaults,
        argument_defaults=argument_defaults,
        prefer_keywords=prefer_keywords,
    )
    assert script == expected



from finesse.cymath.cmatrix cimport SubCCSView, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport conj, cexp
from finesse.simulations.base cimport ModelSettings
from finesse.frequency cimport frequency_info_t
from finesse.simulations.simulation cimport BaseSimulation
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.simulations.workspace cimport GouyFuncWrapper
from finesse.element_workspace cimport BaseCValues

import numpy as np
cimport numpy as np
from cpython.ref cimport PyObject

cdef struct laser_connections:
    # 1D array of SubCCSViews
    PyObject*** SIGPWR_P1o
    PyObject*** SIGAMP_P1o
    PyObject*** SIGFRQ_P1o
    PyObject*** SIGPHS_P1o
    PyObject*** dz_P1o
    PyObject*** dx_P1o
    PyObject*** dy_P1o
    PyObject*** xbeta_P1o
    PyObject*** ybeta_P1o

cdef class LaserConnections:
    cdef public:
        int SIGPWR_P1o_idx
        int SIGAMP_P1o_idx
        int SIGFRQ_P1o_idx
        int SIGPHS_P1o_idx
        int dz_P1o_idx
        int dx_P1o_idx
        int dy_P1o_idx
        int ybeta_P1o_idx
        int xbeta_P1o_idx

    cdef readonly:
        SubCCSView2DArray SIGPWR_P1o
        SubCCSView2DArray SIGAMP_P1o
        SubCCSView2DArray SIGFRQ_P1o
        SubCCSView2DArray SIGPHS_P1o
        SubCCSView2DArray dz_P1o
        SubCCSView2DArray dx_P1o
        SubCCSView2DArray dy_P1o
        SubCCSView2DArray xbeta_P1o
        SubCCSView2DArray ybeta_P1o
    cdef:
        laser_connections ptrs

cdef class LaserValues(BaseCValues):
    cdef public:
        double P
        double phase
        double f
        bint signals_only


cdef class LaserWorkspace(ConnectorWorkspace):
    cdef public:
        Py_ssize_t fsrc_car_idx
        Py_ssize_t fcar_sig_sb_idx[2]
        complex_t[::1] power_coeffs # length sim.model_settings.num_HOMs
        complex_t[::1] hom_vector   # length sim.model_settings.num_HOMs
        complex_t[::1] phase_vector # length sim.model_settings.num_HOMs
        Py_ssize_t node_car_id, node_sig_id
        LaserValues cvalues
        LaserConnections lc
        complex_t PIj_2
        Py_ssize_t P1o_id
        KnmMatrix K_yaw_sig, K_pitch_sig
        bint add_gouy_phase


cdef object c_laser_carrier_fill_rhs(ConnectorWorkspace cws)



"""Higher order function tests."""

from enum import auto, unique, Enum, Flag, IntFlag
import pytest
from finesse.utilities.functools import valuedispatchmethod, flagdispatchmethod


@unique
class ValueField(Enum):
    A = 1
    B = 2
    C = "foo"
    C2 = "bar"


class ValueDispatchNoArgs:
    @valuedispatchmethod
    def get(self, value):
        return value

    @get.register(1)
    def _(self):
        return 1

    @get.register(2)
    def _(self):
        return 2

    @get.register(ValueField.A)
    def _(self):
        return ValueField.A

    @get.register(ValueField.B)
    def _(self):
        return ValueField.B

    @get.register(ValueField.C)
    @get.register(ValueField.C2)
    def _(self):
        return ValueField.C


class ValueDispatchArgs:
    @valuedispatchmethod
    def get(self, value, a, b):
        return value, a, b

    @get.register(1)
    def _(self, a, b):
        return 1, a, b

    @get.register(2)
    def _(self, a, b):
        return 2, a, b

    @get.register(ValueField.A)
    def _(self, a, b):
        return ValueField.A, a, b

    @get.register(ValueField.B)
    def _(self, a, b):
        return ValueField.B, a, b

    @get.register(ValueField.C)
    @get.register(ValueField.C2)
    def _(self, a, b):
        return ValueField.C, a, b


values = pytest.mark.parametrize(
    "value,expected",
    (
        (-1, -1),
        (1, 1),
        (2, 2),
        (ValueField.A, ValueField.A),
        (ValueField.B, ValueField.B),
        (ValueField.C, ValueField.C),
        (ValueField.C2, ValueField.C),
    ),
)


class Flags(Flag):
    A = auto()
    B = auto()
    C = auto()
    C2 = auto()
    D = auto()
    E = auto()
    F = auto()
    UNDEFINED = auto()


class FlagDispatch:
    @flagdispatchmethod
    def get(self, flags):
        return

    @get.register(Flags.A)
    def _(self):
        return Flags.A

    @get.register(Flags.B)
    def _(self):
        return Flags.B

    @get.register(Flags.C)
    @get.register(Flags.C2)
    def _(self):
        return Flags.C

    @get.register(Flags.D | Flags.E)
    @get.register(Flags.F)
    def _(self):
        return Flags.D


flags = pytest.mark.parametrize(
    "flag,expected",
    (
        (Flags.A, Flags.A),
        (Flags.B, Flags.B),
        (Flags.C, Flags.C),
        (Flags.C2, Flags.C),
        (Flags.D, Flags.D),
        (Flags.E, Flags.D),
        (Flags.D | Flags.E, Flags.D),
        (Flags.F, Flags.D),
        (Flags.UNDEFINED, None),
    ),
)


class IntFlags(IntFlag):
    A = auto()
    B = auto()
    C = auto()
    C2 = auto()
    D = auto()
    E = auto()
    F = auto()
    UNDEFINED = auto()


class IntFlagDispatch:
    @flagdispatchmethod
    def get(self, flags):
        return

    @get.register(IntFlags.A)
    def _(self):
        return IntFlags.A

    @get.register(IntFlags.B)
    def _(self):
        return IntFlags.B

    @get.register(IntFlags.C)
    @get.register(IntFlags.C2)
    def _(self):
        return IntFlags.C

    @get.register(IntFlags.D | IntFlags.E)
    @get.register(IntFlags.F)
    def _(self):
        return IntFlags.D

    @get.register(IntFlags.D | IntFlags.F)
    def _(self):
        raise RuntimeError("shouldn't be reached")


intflags = pytest.mark.parametrize(
    "flag,expected",
    (
        (IntFlags.A, IntFlags.A),
        (IntFlags.B, IntFlags.B),
        (IntFlags.C, IntFlags.C),
        (IntFlags.C2, IntFlags.C),
        (IntFlags.D, IntFlags.D),
        (IntFlags.E, IntFlags.D),
        (IntFlags.D | IntFlags.E, IntFlags.D),
        (IntFlags.F, IntFlags.D),
        (IntFlags.UNDEFINED, None),
    ),
)


@values
def test_value_dispatch_noargs(value, expected):
    dispatcher = ValueDispatchNoArgs()
    assert dispatcher.get(value) == expected


@values
def test_value_dispatch_args(value, expected):
    dispatcher = ValueDispatchArgs()
    assert dispatcher.get(value, 1, 2) == (expected, 1, 2)


@flags
def test_flag_dispatch(flag, expected):
    dispatcher = FlagDispatch()
    assert dispatcher.get(flag) == expected


@intflags
def test_int_flag_dispatch(flag, expected):
    dispatcher = IntFlagDispatch()
    assert dispatcher.get(flag) == expected



import pytest
import numpy as np
import finesse.thermal.hello_vinet as hv
from finesse.thermal.hello_vinet import (
    zeros_of_xjn_1_chi_jn,
    thermal_lenses_HG00,
    substrate_temperatures_HG00,
    surface_deformation_coating_heating_HG00,
    surface_deformation_substrate_heating_HG00,
    get_p_n_s_numerical,
    eval_p_n_s_numerical,
)
from scipy.special import jv, eval_hermite
from finesse.materials import FusedSilica
from os import path


@pytest.mark.parametrize("chi", (1e-4, 1, 100))
@pytest.mark.parametrize("n_max", (10,))
@pytest.mark.parametrize("s_max", (10, 50))
def test_zeros(chi, n_max, s_max):
    ns = np.arange(n_max + 1)
    eta = zeros_of_xjn_1_chi_jn(chi, n_max, s_max, 1e-6)
    x = eta
    zeros = x * jv(ns[:, np.newaxis] + 1, x) - chi * jv(ns[:, np.newaxis], x)
    assert abs(zeros).max() < 1e-5


@pytest.mark.parametrize("a", (0.1, 1))
@pytest.mark.parametrize("h", (0.1, 1))
@pytest.mark.parametrize("w", (0.01, 0.05))
@pytest.mark.parametrize("s_max", (10,))
def test_integrate_sub_T_vs_direct(a, h, w, s_max):
    r = np.linspace(0, a, 100)
    z = np.linspace(-h / 2, h / 2, 1000)
    dz = z[1] - z[0]
    material = FusedSilica

    Z_coat_per_W, Z_bulk_per_W = thermal_lenses_HG00(r, a, h, w, material, s_max=s_max)

    T_coat_per_W, T_bulk_per_W = substrate_temperatures_HG00(
        r, z, a, h, w, material, s_max=s_max
    )

    int_coat = dz * material.dndT * (T_coat_per_W).sum(0)
    int_bulk = dz * material.dndT * (T_bulk_per_W).sum(0)

    # loose requirements that the numerical integral is about the same as the direct calculation
    assert (abs(int_coat - Z_coat_per_W) / abs(Z_coat_per_W)).max() < 0.03
    assert (abs(int_bulk - Z_bulk_per_W) / abs(Z_bulk_per_W)).max() < 0.03


def test_coating_heating_thermo_refractive_and_elastic_vs_fea(request):
    """Tests HV gives similar result to FEA."""
    FEA = np.load(
        path.join(
            path.dirname(request.fspath),
            "test_hello_vinet_fea_data_coating_heating.npz",
        )
    )
    R_ap = 0.17
    h = 0.2
    w = 53e-3
    material = FusedSilica
    r = FEA["surface"][:, 0]

    W_z_coat_per_W = thermal_lenses_HG00(
        r, R_ap, h, w, material, s_max=7, root_accuracy=1e-3
    )[0]
    W_z_coat_per_W -= abs(W_z_coat_per_W).max()

    assert abs(W_z_coat_per_W - FEA["substrate"][:, 1]).max() < 15e-9

    Z = -surface_deformation_coating_heating_HG00(
        FEA["surface"][:, 0], R_ap, h, w, material, s_max=7, root_accuracy=1e-3
    )
    Z -= abs(Z).min()

    assert abs(Z - FEA["surface"][:, 1]).max() < 2e-9


def test_substrate_heating_thermo_refractive_and_elastic_vs_fea(request):
    FEA = np.load(
        path.join(
            path.dirname(request.fspath),
            "test_hello_vinet_fea_data_bulk_heating.npz",
        )
    )
    R_ap = 0.17
    h = 0.2
    w = 53e-3
    material = FusedSilica
    r = FEA["r"]

    W_z_bulk_per_W = thermal_lenses_HG00(
        r, R_ap, h, w, material, s_max=7, root_accuracy=1e-3
    )[1]

    assert abs(W_z_bulk_per_W - FEA["substrate"]).max() < 15e-9

    Z = surface_deformation_substrate_heating_HG00(
        r, R_ap, h, w, material, s_max=7, root_accuracy=1e-3
    )

    assert abs(Z - FEA["surface"]).max() < 2e-9


def test_numerical_p_n_s_fit():
    """Tests fourier-bessel decomposition fit."""
    a = 0.17
    w = 53e-3
    r = np.linspace(0, a, 101)
    # 5th order hermite radial distribution
    E = eval_hermite(5, np.sqrt(2) * r / w) * np.exp(-((r / w) ** 2))
    I = E * E
    fit = get_p_n_s_numerical(I, a, 20, FusedSilica)
    assert abs(I - eval_p_n_s_numerical(fit)).max() < 1


def test_substrate_deformation_vs_surface():
    def subtract_min(arr):
        return arr - arr.min()

    w = 53e-3  # beam radius in metres
    a = 170e-3  # radius of optic
    h = 0.2  # thickness of itm

    r = np.linspace(-a, a, 100)  # radial points along optic
    z = np.linspace(-h / 2, +h / 2, 1000)  # depth points along optic
    # @@@ using the full 2d substrate_thermal_expansion_depth_HG00 function @@@
    # 2d array of displacements of the optic due to CP AR1 coating heating
    Uz_method2 = hv.surface_deformation_coating_heating_HG00(
        r, a, h, w, FusedSilica
    ) * (FusedSilica.nr - 1)
    Uz_2d_method1 = hv.substrate_thermal_expansion_depth_HG00(
        r, z, a, h, w, FusedSilica, s_max=20
    ) * (FusedSilica.nr - 1)

    assert np.allclose(subtract_min(Uz_method2), subtract_min(Uz_2d_method1[0, :]))



import sys

try:
    sys.modules[__name__] = __import__("finesse_ligo")
except ImportError:
    raise ImportError(
        "finesse_ligo package cannot be found. To use the LIGO specific tools please install the finesse_ligo package: pip install finesse_ligo"
    )



"""Tests ensuring that data points of a simulation scan where cavities become unstable
correctly transition to a contingent forest mode if any other stable dependency is in
the model."""

import warnings
import numpy as np
import pytest
from finesse import Model
from finesse.analysis.actions import Xaxis
from finesse.components import Gauss
from finesse.gaussian import BeamParam
from finesse.warnings import CavityUnstableWarning


@pytest.fixture
def fp_cavity_model_with_gauss():
    IFO = Model()
    IFO.parse(
        """
        l L0 P=1
        s s0 L0.p1 ITM.p1

        m ITM Rc=-2
        s sc ITM.p2 ETM.p1 L=1
        m ETM Rc=2

        cp cav_q FP q
        # Add a detector relying on internal FP trace tree
        bp fp_q ITM.p2.o q
        # and another relying on an external FP trace tree
        bp in_w L0.p1.o w

        # Add a detector for the g-factor, the outputs of
        # this should never be masked
        cp fp_g FP g

        cav FP ITM.p2
        """
    )
    IFO.ITM.Rcy = IFO.ITM.Rcx.ref
    IFO.ETM.Rcy = IFO.ETM.Rcx.ref

    # Make sure output values of fp_q detector are BeamParam objects
    IFO.fp_q.q_as_bp = True

    # Add an input beam with q matched to cavity
    trace = IFO.beam_trace()
    IFO.add(Gauss("gL0", IFO.L0.p1.o, q=trace[IFO.L0.p1.o]))

    return IFO


def test_critically_stable_fp(fp_cavity_model_with_gauss: Model):
    IFO = fp_cavity_model_with_gauss

    # Ignore the unstable cavity warning for the purposes of this test.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=CavityUnstableWarning)
        # Scan lengths [1.9, 2.0, 2.1] -> 2 should give g = 0
        out = IFO.run(Xaxis("sc.L", "lin", 1.9, 2.1, 2))

    q_modes = out["cav_q"]
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    # FP has entered critically stable point but gL0 still exists
    # so the simulation should use the tree from gL0 as a contingent
    # forest now --- meaning nothing should be masked
    assert not np.ma.is_masked(qs)
    assert not np.ma.is_masked(ws)
    assert not np.ma.is_masked(gs)

    assert gs[1] == 0

    gauss_q = BeamParam(q=IFO.gL0.qx)

    assert qs[0] != gauss_q
    assert qs[2] != gauss_q
    # q at cavity node should be equal to gauss q (as L=0 between L0 and ITM)
    # for the point where the cavity becomes critically stable
    assert qs[1] == gauss_q

    # Cavity eigenmode at g = 0 point should be NaN
    assert np.isnan(q_modes[1])


def test_unstable_fp(fp_cavity_model_with_gauss: Model):
    IFO = fp_cavity_model_with_gauss

    # Ignore the unstable cavity warning for the purposes of this test.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=CavityUnstableWarning)
        # Scan lengths [3.9, 4.0, 4.1] -> 4 should give g = 1, 4.1 should give g > 1
        out = IFO.run(Xaxis("sc.L", "lin", 3.9, 4.1, 2))

    q_modes = out["cav_q"]
    qs = out["fp_q"]
    ws = out["in_w"]
    gs = out["fp_g"]

    # FP has entered critically stable point but gL0 still exists
    # so the simulation should use the tree from gL0 as a contingent
    # forest now --- meaning nothing should be masked
    assert not np.ma.is_masked(qs)
    assert not np.ma.is_masked(ws)
    assert not np.ma.is_masked(gs)

    assert gs[1] == 1
    assert gs[2] > 1

    gauss_q = BeamParam(q=IFO.gL0.qx)

    assert qs[0] != gauss_q
    # q at cavity node should be equal to gauss q (as L=0 between L0 and ITM)
    # for g = 1 and g > 1
    assert qs[1] == gauss_q
    assert qs[2] == gauss_q

    # Cavity eigenmode at g >= 1 points should be NaN
    assert np.isnan(q_modes[1])
    assert np.isnan(q_modes[2])


# TODO (sjr) More tests for other configurations, especially those which include
#            coupled / overlapping cavities



from .base import Action
from .series import Series, For
from .parallel import Parallel
from .sweep import Sweep
from .beam import ABCD, BeamTrace, PropagateAstigmaticBeam, PropagateBeam
from .axes import Noxaxis, Xaxis, X2axis, X3axis
from .random import (
    Change,
    UpdateMaps,
    Plot,
    Printer,
    PrintModel,
    PrintModelAttr,
    StoreModelAttr,
    Execute,
)
from .debug import Debug
from .lti import (
    FrequencyResponse,
    FrequencyResponse2,
    FrequencyResponse3,
    FrequencyResponse4,
)
from .sensing import (
    OptimiseRFReadoutPhaseDC,
    SensingMatrixDC,
    SensingMatrixAC,
    GetErrorSignals,
    CheckLinearity,
)
from .operator import Operator, Eigenmodes
from .locks import RunLocks, DragLocks, SetLockGains
from .noise import NoiseProjection
from .optimisation import Maximize, Minimize, Optimize
from .temporary import Temporary, TemporaryParameters
from .pseudolock import PseudoLockCavity, PseudoLockDRFPMI
from .squeezing import AntiSqueezing
from .dc import DCFields


__all__ = (
    "Action",
    "Series",
    "For",
    "Parallel",
    "Sweep",
    "ABCD",
    "BeamTrace",
    "PropagateBeam",
    "PropagateAstigmaticBeam",
    "Xaxis",
    "X2axis",
    "X3axis",
    "Noxaxis",
    "Debug",
    "Plot",
    "Change",
    "Printer",
    "PrintModel",
    "PrintModelAttr",
    "UpdateMaps",
    "FrequencyResponse",
    "FrequencyResponse2",
    "FrequencyResponse3",
    "FrequencyResponse4",
    "OptimiseRFReadoutPhaseDC",
    "CheckLinearity",
    "SensingMatrixDC",
    "SensingMatrixAC",
    "GetErrorSignals",
    "Operator",
    "Eigenmodes",
    "RunLocks",
    "NoiseProjection",
    "DragLocks",
    "Minimize",
    "Maximize",
    "Temporary",
    "TemporaryParameters",
    "StoreModelAttr",
    "PseudoLockCavity",
    "PseudoLockDRFPMI",
    "Optimize",
    "Execute",
    "AntiSqueezing",
    "SetLockGains",
    "DCFields",
)



""""Finesse object comparison utilities for tests.

There is a lot going on in this module. We cannot compare some Finesse objects by
implementing `__eq__` methods directly (doing so would interfere with how they behave in
lists, dicts, etc.), so special checks are instead implemented here. Furthermore, in
order to provide useful error messages for tests, special set operations are provided
that use the special equivalence functions here instead of hashing.
"""

from io import BytesIO

import networkx as nx
import numpy as np
from numpy.testing import assert_array_equal

from finesse.components import Cavity, Gauss
from finesse.detectors.general import Detector


def assert_models_equivalent(model_a, model_b):
    """Check :class:`.Model` equality.

    This checks that the two specified models are identical in a shallow way, i.e. they
    have elements with the same properties, equivalent networks, actions, etc.

    This is implemented here and not as `Model.__eq__` because :class:`.Model` must be
    hashable.
    """
    if model_a.analysis is not None and model_b.analysis is not None:
        assert_actions_equivalent(model_a.analysis, model_b.analysis)
    elif (model_a.analysis is None) != (model_b.analysis is None):
        raise AssertionError("analyses don't match")

    assert set(model_a.elements) == set(model_b.elements)
    assert_array_equal(model_a.homs, model_b.homs, "HOMs don't match")

    for el_a, el_b in zip(sorted(model_a.elements), sorted(model_b.elements)):
        assert_model_elements_equivalent(model_a.elements[el_a], model_b.elements[el_b])


def assert_model_elements_equivalent(element_a, element_b):
    """Check :class:`.ModelElement` equality.

    This is implemented here and not as `ModelElement.__eq__` because `ModelElement`
    must be hashable.
    """

    assert type(element_a) == type(element_b)  # noqa: E721
    assert element_a.name == element_b.name
    for param_a, param_b in zip(element_a.parameters, element_b.parameters):
        assert_parameters_equivalent(param_a, param_b)


def assert_nodes_equivalent(node_a, node_b):
    """Check node equality.

    This is implemented here and not as `Node.__eq__` because `Node` must be hashable.
    """
    assert node_a.name == node_b.name, f"{node_a!r} name != {node_b!r} name"
    assert_model_elements_equivalent(node_a.component, node_b.component)


def assert_parameters_equivalent(param_a, param_b, expected=None):
    """Check model parameter equality.

    This is implemented here and not as `ModelParameter.__eq__` because `ModelParameter`
    must be hashable.
    """
    assert type(param_a) == type(param_b)  # noqa:E721

    val_a = param_a.eval()
    val_b = param_b.eval()

    assert val_a == val_b

    if expected is not None:
        assert val_a == expected
        assert val_b == expected


# FIXME: make this check far more than just the name of the top action...
def assert_actions_equivalent(act_a, act_b):
    """Check action equality."""
    # TODO: compare more things.
    assert str(act_a.plan()) == str(act_b.plan())


def assert_cavities_equivalent(cav_a, cav_b):
    """Check cavity equality."""
    if not isinstance(cav_b, cav_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(cav_a, cav_b)
    assert isinstance(cav_a, Cavity), f"{cav_a!r} is not a Cavity"
    assert isinstance(cav_b, Cavity), f"{cav_b!r} is not a Cavity"
    assert (
        cav_a.round_trip_optical_length == cav_b.round_trip_optical_length
    ), f"{cav_a!r} round trip length != {cav_b!r} round trip length"
    assert np.all(
        cav_a.gouy == cav_b.gouy
    ), f"{cav_a!r} round trip gouy != {cav_b!r} round trip gouy"
    assert cav_a.FSR == cav_b.FSR, f"{cav_a!r} FSR != {cav_b!r} FSR"
    assert cav_a.loss == cav_b.loss, f"{cav_a!r} loss != {cav_b!r} loss"
    assert cav_a.finesse == cav_b.finesse, f"{cav_a!r} finesse != {cav_b!r} finesse"
    assert cav_a.FWHM == cav_b.FWHM, f"{cav_a!r} FWHM != {cav_b!r} FWHM"
    assert cav_a.pole == cav_b.pole, f"{cav_a!r} pole != {cav_b!r} pole"
    assert np.all(
        cav_a.mode_separation == cav_b.mode_separation
    ), f"{cav_a!r} mode separation != {cav_b!r} mode separation"
    assert np.all(cav_a.S == cav_b.S), f"{cav_a!r} resolution != {cav_b!r} resolution"
    assert np.all(cav_a.ABCDx == cav_b.ABCDx), f"{cav_a!r} ABCDx != {cav_b!r} ABCDx"
    assert np.all(cav_a.ABCDy == cav_b.ABCDy), f"{cav_a!r} ABCDy != {cav_b!r} ABCDy"
    assert np.all(cav_a.q == cav_b.q), f"{cav_a!r} q != {cav_b!r} q"
    assert np.all(cav_a.g == cav_b.g), f"{cav_a!r} g != {cav_b!r} g"


def assert_gauss_commands_equivalent(gauss_a, gauss_b):
    """Check gauss command equality.

    This is implemented here and not as `Gauss.__eq__` because `Gauss` must be hashable.
    """
    if not isinstance(gauss_b, gauss_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(gauss_a, gauss_b)
    assert isinstance(gauss_a, Gauss), f"{gauss_a!r} is not a Gauss"
    assert isinstance(gauss_b, Gauss), f"{gauss_b!r} is not a Gauss"
    # TODO: add more gauss command comparisons.


def assert_detectors_equivalent(det_a, det_b):
    """Check detector equality.

    This is implemented here and not as `Detector.__eq__` because `Detector` must be
    hashable.
    """
    if not isinstance(det_b, det_a.__class__):
        # Don't attempt to compare against unrelated types.
        raise NotImplementedError

    assert_model_elements_equivalent(det_a, det_b)
    assert isinstance(det_a, Detector), f"{det_a!r} is not a Detector"
    assert isinstance(det_b, Detector), f"{det_b!r} is not a Detector"
    # TODO: add more detector comparisons.


def assert_trace_trees_equal(t1, t2):
    """Check TraceTree equality."""
    # Both t1 and t2 must be None or not None.
    if t1 is None and t2 is None:
        # Nothing to test.
        return
    elif (t1 is None and t2 is not None) or (t1 is not None and t2 is None):
        raise AssertionError

    assert t1.node == t2.node, f"{t1} node {t1.node} != {t2} node {t2.node}"
    assert (
        t1.dependency == t2.dependency
    ), f"{t1} dependency {t1.dependency} != {t2} dependency {t2.dependency}"

    assert_trace_trees_equal(t1.left, t2.left)
    assert_trace_trees_equal(t1.right, t2.right)


def assert_graph_matches_def(graph, ref_definition):
    """Assert specified graph has GML representation equivalent to specified
    definition."""
    ref_graph = nx.read_gml(BytesIO(ref_definition.encode()))
    assert_graphs_match(graph, ref_graph)


def assert_graphs_match(graph, ref_graph):
    """Assert graphs match."""

    def stringify(item):
        """Convert graph item to string form."""
        if isinstance(item, list):
            return f"[{', '.join(stringify(i) for i in item)}]"
        return str(item)

    nodes = sorted([node for node in graph.nodes], key=lambda node: str(node))
    ref_nodes = sorted([node for node in ref_graph.nodes], key=lambda node: str(node))

    # Node names.
    assert [stringify(node) for node in nodes] == [
        stringify(node) for node in ref_nodes
    ], "Node names in graph don't match those in reference graph"

    # Node attributes.
    for node, ref_node in zip(nodes, ref_nodes):
        data = {
            stringify(key): stringify(value) for key, value in graph.nodes[node].items()
        }
        ref_data = {
            stringify(key): stringify(value)
            for key, value in ref_graph.nodes[ref_node].items()
        }
        assert (
            data == ref_data
        ), f"{node!r} attributes in graph don't match those in reference graph"

    edges = sorted(
        [edge for edge in graph.edges],
        key=lambda edge: (stringify(edge[0]), stringify(edge[1])),
    )
    ref_edges = sorted(
        [edge for edge in ref_graph.edges],
        key=lambda edge: (stringify(edge[0]), stringify(edge[1])),
    )

    # Edges.
    assert list(edges) == list(ref_edges)

    # Edge attributes.
    for edge, ref_edge in zip(edges, ref_edges):
        data = {
            stringify(key): stringify(value) for key, value in graph.edges[edge].items()
        }
        ref_data = {
            stringify(key): stringify(value)
            for key, value in ref_graph.edges[ref_edge].items()
        }
        assert (
            data == ref_data
        ), f"{edge!r} attributes in graph don't match those in reference graph"
